#!/usr/bin/env python

"""
Tests for saving a UGrid in netcdf format.

Designed to be run with pytest.

"""

from __future__ import (absolute_import, division, print_function)

import os
import netCDF4
import numpy as np

from pyugrid.ugrid import UGrid, UVar

from .utilities import chdir, two_triangles, twenty_one_triangles

test_files = os.path.join(os.path.dirname(__file__), 'files')


def nc_has_variable(ds, var_name):
    """
    Checks that a netcdf file has the given variable defined.

    :param ds: a netCDF4.Dataset object, or a netcdf file name.

    """
    if not isinstance(ds, netCDF4.Dataset):
        ds = netCDF4.Dataset(ds)

    if var_name in ds.variables:
        return True
    else:
        print('{} is not a variable in the Dataset'.format(var_name))
        return False


def nc_has_dimension(ds, dim_name):
    """
    Checks that a netcdf file has the given dimension defined.

    :param ds: a netCDF4.Dataset object, or a netcdf file name.

    """
    if not isinstance(ds, netCDF4.Dataset):
        ds = netCDF4.Dataset(ds)
    if dim_name in ds.dimensions:
        return True
    else:
        return False


def nc_var_has_attr(ds, var_name, att_name):
    """
    Checks that the variable, var_name, has the attribute, att_name.

    """
    if not isinstance(ds, netCDF4.Dataset):
        ds = netCDF4.Dataset(ds)

    try:
        getattr(ds.variables[var_name], att_name)
        return True
    except AttributeError:
        return False


def nc_var_has_attr_vals(ds, var_name, att_dict):
    """
    Checks that the variable, var_name, as the attributes (and values)
    in the att_dict.

    """
    if not isinstance(ds, netCDF4.Dataset):
        ds = netCDF4.Dataset(ds)

    for key, val in att_dict.items():
        try:
            if val != getattr(ds.variables[var_name], key):
                return False
        except AttributeError:
            return False
    return True


def test_simple_write():
    grid = two_triangles()

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    # TODO: Could be lots of tests here.
    assert nc_has_variable(ds, 'mesh')
    assert nc_var_has_attr_vals(ds, 'mesh', {
        'cf_role': 'mesh_topology',
        'topology_dimension': 2,
        'long_name': u'Topology data of 2D unstructured mesh'})
    ds.close()


def test_set_mesh_name():
    grid = two_triangles()
    grid.mesh_name = 'mesh_2'

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    assert nc_has_variable(ds, 'mesh_2')
    assert nc_var_has_attr_vals(ds, 'mesh_2', {
        'cf_role': 'mesh_topology',
        'topology_dimension': 2,
        'long_name': u'Topology data of 2D unstructured mesh'})
    assert nc_var_has_attr_vals(ds, 'mesh_2', {
        'cf_role': 'mesh_topology',
        'topology_dimension': 2,
        'long_name': u'Topology data of 2D unstructured mesh',
        'node_coordinates': 'mesh_2_node_lon mesh_2_node_lat'})

    assert nc_has_variable(ds, 'mesh_2_node_lon')
    assert nc_has_variable(ds, 'mesh_2_node_lat')
    assert nc_has_variable(ds, 'mesh_2_face_nodes')
    assert nc_has_variable(ds, 'mesh_2_edge_nodes')

    assert nc_has_dimension(ds, 'mesh_2_num_node')
    assert nc_has_dimension(ds, 'mesh_2_num_edge')
    assert nc_has_dimension(ds, 'mesh_2_num_face')
    assert nc_has_dimension(ds, 'mesh_2_num_vertices')

    assert not nc_var_has_attr(ds, 'mesh_2', 'face_edge_connectivity')
    ds.close()


def test_write_with_depths():
    """Tests writing a netcdf file with depth data."""

    grid = two_triangles()
    grid.mesh_name = 'mesh1'

    # Create a UVar object for the depths:
    depths = UVar('depth', location='node', data=[1.0, 2.0, 3.0, 4.0])
    depths.attributes['units'] = 'm'
    depths.attributes['standard_name'] = 'sea_floor_depth_below_geoid'
    depths.attributes['positive'] = 'down'

    grid.add_data(depths)

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    assert nc_has_variable(ds, 'mesh1')
    assert nc_has_variable(ds, 'depth')
    assert nc_var_has_attr_vals(ds, 'depth', {
        'coordinates': 'mesh1_node_lon mesh1_node_lat',
        'location': 'node',
        'mesh': 'mesh1'})
    ds.close()


def test_write_with_velocities():
    """Tests writing a netcdf file with velocities on the faces."""
    grid = two_triangles()
    grid.mesh_name = 'mesh2'

    # Create a UVar object for u velocity:
    u_vel = UVar('u', location='face', data=[1.0, 2.0])
    u_vel.attributes['units'] = 'm/s'
    u_vel.attributes['standard_name'] = 'eastward_sea_water_velocity'

    grid.add_data(u_vel)

    # Create a UVar object for v velocity:
    v_vel = UVar('v', location='face', data=[3.2, 4.3])
    v_vel.attributes['units'] = 'm/s'
    v_vel.attributes['standard_name'] = 'northward_sea_water_velocity'

    grid.add_data(v_vel)

    # Add coordinates for face data.
    grid.build_face_coordinates()

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    assert nc_has_variable(ds, 'mesh2')
    assert nc_has_variable(ds, 'u')
    assert nc_has_variable(ds, 'v')
    assert nc_var_has_attr_vals(ds, 'u', {
        'coordinates': 'mesh2_face_lon mesh2_face_lat',
        'location': 'face',
        'mesh': 'mesh2',
        })
    ds.close()


def test_write_with_edge_data():
    """Tests writing a netcdf file with data on the edges (fluxes, maybe?)."""

    grid = two_triangles()
    grid.mesh_name = 'mesh2'

    # Create a UVar object for fluxes:
    flux = UVar('flux', location='edge', data=[0.0, 0.0, 4.1, 0.0, 5.1, ])
    flux.attributes['units'] = 'm^3/s'
    flux.attributes['long_name'] = 'volume flux between cells'
    flux.attributes['standard_name'] = 'ocean_volume_transport_across_line'

    grid.add_data(flux)

    # Add coordinates for edges.
    grid.build_edge_coordinates()

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    assert nc_has_variable(ds, 'mesh2')
    assert nc_has_variable(ds, 'flux')
    assert nc_var_has_attr_vals(ds, 'flux', {
        'coordinates': 'mesh2_edge_lon mesh2_edge_lat',
        'location': 'edge',
        'units': 'm^3/s',
        'mesh': 'mesh2'})
    assert np.array_equal(ds.variables['mesh2_edge_lon'],
                          grid.edge_coordinates[:, 0])
    assert np.array_equal(ds.variables['mesh2_edge_lat'],
                          grid.edge_coordinates[:, 1])
    ds.close()


def test_write_with_bound_data():
    """
    Tests writing a netcdf file with data on the boundaries
    suitable for boundary conditions, for example fluxes.

    """
    grid = two_triangles()

    # Add the boundary definitions:
    grid.boundaries = [(0, 1),
                       (0, 2),
                       (1, 3),
                       (2, 3)]

    # Create a UVar object for boundary conditions:
    bnds = UVar('bnd_cond', location='boundary', data=[0, 1, 0, 0])
    bnds.attributes['long_name'] = 'model boundary conditions'
    bnds.attributes['flag_values'] = '0 1'
    bnds.attributes['flag_meanings'] = 'no_flow_boundary  open_boundary'

    grid.add_data(bnds)

    fname = 'temp.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)
        os.remove(fname)

    assert nc_has_variable(ds, 'mesh')
    assert nc_has_variable(ds, 'bnd_cond')
    assert nc_var_has_attr_vals(ds, 'mesh', {
        'boundary_node_connectivity': 'mesh_boundary_nodes'})
    assert nc_var_has_attr_vals(ds, 'bnd_cond', {
        'location': 'boundary',
        'flag_values': '0 1',
        'flag_meanings': 'no_flow_boundary  open_boundary',
        'mesh': 'mesh',
        })
    # There should be no coordinates attribute or variable for the
    # boundaries as there is no boundaries_coordinates defined.
    assert not nc_has_variable(ds, 'mesh_boundary_lon')
    assert not nc_has_variable(ds, 'mesh_boundary_lat')
    assert not nc_var_has_attr(ds, 'bnd_cond', 'coordinates')
    ds.close()


def test_write_everything():
    """An example with all features enabled, and a less trivial grid."""

    grid = twenty_one_triangles()

    grid.build_edges()
    grid.build_face_face_connectivity()

    grid.build_edge_coordinates()
    grid.build_face_coordinates()
    grid.build_boundary_coordinates()

    # Depth on the nodes.
    depths = UVar('depth', location='node', data=np.linspace(1, 10, 20))
    depths.attributes['units'] = 'm'
    depths.attributes['standard_name'] = 'sea_floor_depth_below_geoid'
    depths.attributes['positive'] = 'down'
    grid.add_data(depths)

    # Create a UVar object for u velocity:
    u_vel = UVar('u', location='face', data=np.sin(np.linspace(3, 12, 21)))
    u_vel.attributes['units'] = 'm/s'
    u_vel.attributes['standard_name'] = 'eastward_sea_water_velocity'

    grid.add_data(u_vel)

    # Create a UVar object for v velocity:
    v_vel = UVar('v', location='face', data=np.sin(np.linspace(12, 15, 21)))
    v_vel.attributes['units'] = 'm/s'
    v_vel.attributes['standard_name'] = 'northward_sea_water_velocity'

    grid.add_data(v_vel)

    # Fluxes on the edges:
    flux = UVar('flux', location='edge', data=np.linspace(1000, 2000, 41))
    flux.attributes['units'] = 'm^3/s'
    flux.attributes['long_name'] = 'volume flux between cells'
    flux.attributes['standard_name'] = 'ocean_volume_transport_across_line'

    grid.add_data(flux)

    # Boundary conditions:
    bounds = np.zeros((19,), dtype=np.uint8)
    bounds[7] = 1
    bnds = UVar('bnd_cond', location='boundary', data=bounds)
    bnds.attributes['long_name'] = 'model boundary conditions'
    bnds.attributes['flag_values'] = '0 1'
    bnds.attributes['flag_meanings'] = 'no_flow_boundary  open_boundary'

    grid.add_data(bnds)

    fname = 'full_example.nc'
    with chdir(test_files):
        grid.save_as_netcdf(fname)
        ds = netCDF4.Dataset(fname)

        # Now the tests:
        assert nc_has_variable(ds, 'mesh')
        assert nc_has_variable(ds, 'depth')
        assert nc_var_has_attr_vals(ds, 'depth', {
            'coordinates': 'mesh_node_lon mesh_node_lat',
            'location': 'node'})
        assert nc_has_variable(ds, 'u')
        assert nc_has_variable(ds, 'v')
        assert nc_var_has_attr_vals(ds, 'u', {
            'coordinates': 'mesh_face_lon mesh_face_lat',
            'location': 'face',
            'mesh': 'mesh'})
        assert nc_var_has_attr_vals(ds, 'v', {
            'coordinates': 'mesh_face_lon mesh_face_lat',
            'location': 'face',
            'mesh': 'mesh',
            })
        assert nc_has_variable(ds, 'flux')
        assert nc_var_has_attr_vals(ds, 'flux', {
            'coordinates': 'mesh_edge_lon mesh_edge_lat',
            'location': 'edge',
            'units': 'm^3/s',
            'mesh': 'mesh'})
        assert nc_has_variable(ds, 'mesh')
        assert nc_has_variable(ds, 'bnd_cond')
        assert nc_var_has_attr_vals(ds, 'mesh', {
            'boundary_node_connectivity': 'mesh_boundary_nodes',
            })
        assert nc_var_has_attr_vals(ds, 'bnd_cond', {
            'location': 'boundary',
            'flag_values': '0 1',
            'flag_meanings': 'no_flow_boundary  open_boundary',
            'mesh': 'mesh',
            })
        ds.close()

        # And make sure pyugrid can reload it!
        with chdir(test_files):
            grid = UGrid.from_ncfile('full_example.nc', load_data=True)
        # And that some things are the same.
        # NOTE: more testing might be good here.
        # maybe some grid comparison functions?

        assert grid.mesh_name == 'mesh'
        assert len(grid.nodes) == 20

        depth = grid.data['depth']
        assert depth.attributes['units'] == 'm'

        u = grid.data['u']
        assert u.attributes['units'] == 'm/s'

        os.remove(fname)


if __name__ == "__main__":
    test_simple_write()
    test_set_mesh_name()
    test_write_with_depths()
    test_write_with_velocities()
    test_write_with_edge_data()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/01/2005
#
#------------------------------------------------------------------------------

""" Defines the table column descriptor used by the editor and editor factory
    classes for numeric and table editors.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import, print_function

import os

from traits.api import (
    Any,
    Bool,
    Callable,
    Color,
    Constant,
    Either,
    Enum,
    Expression,
    Float,
    Font,
    HasPrivateTraits,
    Instance,
    Int,
    Property,
    Str)

from traits.trait_base import user_name_for, xgetattr

from .editor_factory import EditorFactory
from .menu import Menu
from .ui_traits import Image, AView, EditorStyle
from .view import View

# Set up a logger:
import logging
logger = logging.getLogger(__name__)


TRAITS_DEBUG = (os.getenv('TRAITS_DEBUG') is not None)

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# Flag used to indicate user has not specified a column label
UndefinedLabel = '???'

#-------------------------------------------------------------------------
#  'TableColumn' class:
#-------------------------------------------------------------------------


class TableColumn(HasPrivateTraits):
    """ Represents a column in a table editor.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Column label to use for this column:
    label = Str(UndefinedLabel)

    # Type of data contained by the column:
    type = Enum('text')

    # Text color for this column:
    text_color = Color('black')

    # Text font for this column:
    text_font = Font

    # Cell background color for this column:
    cell_color = Either(Color('white'), None)

    # Cell background color for non-editable columns:
    read_only_cell_color = Either(Color(0xF4F3EE), None)

    # Cell graph color:
    graph_color = Color(0xDDD9CC)

    # Horizontal alignment of text in the column:
    horizontal_alignment = Enum('left', ['left', 'center', 'right'])

    # Vertical alignment of text in the column:
    vertical_alignment = Enum('center', ['top', 'center', 'bottom'])

    # Horizontal cell margin
    horizontal_margin = Int(4)

    # Vertical cell margin
    vertical_margin = Int(3)

    # The image to display in the cell:
    image = Image

    # Renderer used to render the contents of this column:
    renderer = Any  # A toolkit specific renderer

    # Is the table column visible (i.e., viewable)?
    visible = Bool(True)

    # Is this column editable?
    editable = Bool(True)

    # Is the column automatically edited/viewed (i.e. should the column editor
    # or popup be activated automatically on mouse over)?
    auto_editable = Bool(False)

    # Should a checkbox be displayed instead of True/False?
    show_checkbox = Bool(True)

    # Can external objects be dropped on the column?
    droppable = Bool(False)

    # Context menu to display when this column is right-clicked:
    menu = Instance(Menu)

    # The tooltip to display when the mouse is over the column:
    tooltip = Str

    # The width of the column (< 0.0: Default, 0.0..1.0: fraction of total table
    # width, > 1.0: absolute width in pixels):
    width = Float(-1.0)

    # The width of the column while it is being edited (< 0.0: Default,
    # 0.0..1.0: fraction of total table width, > 1.0: absolute width in
    # pixels):
    edit_width = Float(-1.0)

    # The height of the column cell's row while it is being edited
    # (< 0.0: Default, 0.0..1.0: fraction of total table height,
    # > 1.0: absolute height in pixels):
    edit_height = Float(-1.0)

    # The resize mode for this column.  This takes precedence over other settings
    # (like **width**, above).
    #   "interactive": column can be resized by users or programmatically
    #   "fixed": users cannot resize the column, but it can be set programmatically
    #   "stretch": the column will be resized to fill the available space
    #   "resize_to_contents": column will be sized to fit the contents, but then cannot be resized
    resize_mode = Enum("interactive", "fixed", "stretch", "resize_to_contents")

    # The view (if any) to display when clicking a non-editable cell:
    view = AView

    # Optional maximum value a numeric cell value can have:
    maximum = Float(trait_value=True)

    #-------------------------------------------------------------------------
    #  Returns the actual object being edited:
    #-------------------------------------------------------------------------

    def get_object(self, object):
        """ Returns the actual object being edited.
        """
        return object

    #-------------------------------------------------------------------------
    #  Gets the label of the column:
    #-------------------------------------------------------------------------

    def get_label(self):
        """ Gets the label of the column.
        """
        return self.label

    #-------------------------------------------------------------------------
    #  Returns the width of the column:
    #-------------------------------------------------------------------------

    def get_width(self):
        """ Returns the width of the column.
        """
        return self.width

    #-------------------------------------------------------------------------
    #  Returns the edit width of the column:
    #-------------------------------------------------------------------------

    def get_edit_width(self, object):
        """ Returns the edit width of the column.
        """
        return self.edit_width

    #-------------------------------------------------------------------------
    #  Returns the height of the column cell's row while it is being edited:
    #-------------------------------------------------------------------------

    def get_edit_height(self, object):
        """ Returns the height of the column cell's row while it is being
            edited.
        """
        return self.edit_height

    #-------------------------------------------------------------------------
    #  Gets the type of data for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_type(self, object):
        """ Gets the type of data for the column for a specified object.
        """
        return self.type

    #-------------------------------------------------------------------------
    #  Returns the text color for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_text_color(self, object):
        """ Returns the text color for the column for a specified object.
        """
        return self.text_color_

    #-------------------------------------------------------------------------
    #  Returns the text font for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_text_font(self, object):
        """ Returns the text font for the column for a specified object.
        """
        return self.text_font

    #-------------------------------------------------------------------------
    #  Returns the cell background color for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object.
        """
        if self.is_editable(object):
            return self.cell_color_
        return self.read_only_cell_color_

    #-------------------------------------------------------------------------
    #  Returns the cell background graph color for the column for a specified
    #  object:
    #-------------------------------------------------------------------------

    def get_graph_color(self, object):
        """ Returns the cell background graph color for the column for a
            specified object.
        """
        return self.graph_color_

    #-------------------------------------------------------------------------
    #  Returns the horizontal alignment for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_horizontal_alignment(self, object):
        """ Returns the horizontal alignment for the column for a specified
            object.
        """
        return self.horizontal_alignment

    #-------------------------------------------------------------------------
    #  Returns the vertical alignment for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_vertical_alignment(self, object):
        """ Returns the vertical alignment for the column for a specified
            object.
        """
        return self.vertical_alignment

    #-------------------------------------------------------------------------
    #  Returns the image to display for the column for a specified object:
    #-------------------------------------------------------------------------

    def get_image(self, object):
        """ Returns the image to display for the column for a specified object.
        """
        return self.image

    #-------------------------------------------------------------------------
    #  Returns the renderer for the column of a specified object:
    #-------------------------------------------------------------------------

    def get_renderer(self, object):
        """ Returns the renderer for the column of a specified object.
        """
        return self.renderer

    #-------------------------------------------------------------------------
    #  Returns whether the column is editable for a specified object:
    #-------------------------------------------------------------------------

    def is_editable(self, object):
        """ Returns whether the column is editable for a specified object.
        """
        return self.editable

    #-------------------------------------------------------------------------
    #  Returns whether the column is autoamtically edited/viewed for a specified
    #  object:
    #-------------------------------------------------------------------------

    def is_auto_editable(self, object):
        """ Returns whether the column is automatically edited/viewed for a
            specified object.
        """
        return self.auto_editable

    #-------------------------------------------------------------------------
    #  Returns whether a specified value is valid for dropping on the column
    #  for a specified object:
    #-------------------------------------------------------------------------

    def is_droppable(self, object, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object.
        """
        return self.droppable

    #-------------------------------------------------------------------------
    #  Returns the context menu to display when the user right-clicks on the
    #  column for a specified object:
    #-------------------------------------------------------------------------

    def get_menu(self, object):
        """ Returns the context menu to display when the user right-clicks on
            the column for a specified object.
        """
        return self.menu

    #-------------------------------------------------------------------------
    #  Returns the tooltip to display when the user mouses over the column for
    #  a specified object:
    #-------------------------------------------------------------------------

    def get_tooltip(self, object):
        """ Returns the tooltip to display when the user mouses over the column
            for a specified object.
        """
        return self.tooltip

    #-------------------------------------------------------------------------
    #  Returns the view to display when clicking a non-editable cell:
    #-------------------------------------------------------------------------

    def get_view(self, object):
        """ Returns the view to display when clicking a non-editable cell.
        """
        return self.view

    #-------------------------------------------------------------------------
    #  Returns the maximum value a numeric column can have:
    #-------------------------------------------------------------------------

    def get_maximum(self, object):
        """ Returns the maximum value a numeric column can have.
        """
        return self.maximum

    #-------------------------------------------------------------------------
    #  Called when the user clicks on the column:
    #-------------------------------------------------------------------------

    def on_click(self, object):
        """ Called when the user clicks on the column.
        """
        pass

    #-------------------------------------------------------------------------
    #  Called when the user double-clicks on the column:
    #-------------------------------------------------------------------------

    def on_dclick(self, object):
        """ Called when the user clicks on the column.
        """
        pass

    #-------------------------------------------------------------------------
    #  Returns the result of comparing the column of two different objects:
    #-------------------------------------------------------------------------

    def cmp(self, object1, object2):
        """ Returns the result of comparing the column of two different objects.

        This is deprecated.
        """
        return ((self.key(object1) > self.key(object2)) -
                (self.key(object1) < self.key(object2)))

    #-------------------------------------------------------------------------
    #  Returns the string representation of the table column:
    #-------------------------------------------------------------------------

    def __str__(self):
        """ Returns the string representation of the table column.
        """
        return self.get_label()

#-------------------------------------------------------------------------
#  'ObjectColumn' class:
#-------------------------------------------------------------------------


class ObjectColumn(TableColumn):
    """ A column for editing objects.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Name of the object trait associated with this column:
    name = Str

    # Column label to use for this column:
    label = Property

    # Trait editor used to edit the contents of this column:
    editor = Instance(EditorFactory)

    # The editor style to use to edit the contents of this column:
    style = EditorStyle

    # Format string to apply to column values:
    format = Str('%s')

    # Format function to apply to column values:
    format_func = Callable

    #-------------------------------------------------------------------------
    #  Trait view definitions:
    #-------------------------------------------------------------------------

    traits_view = View([['name', 'label', 'type',
                         '|[Column Information]'],
                        ['horizontal_alignment{Horizontal}@',
                         'vertical_alignment{Vertical}@',
                         '|[Alignment]'],
                        ['editable', '9', 'droppable', '9', 'visible',
                         '-[Options]>'],
                        '|{Column}'],
                       [['text_color@', 'cell_color@',
                         'read_only_cell_color@',
                         '|[UI Colors]'],
                        '|{Colors}'],
                       [['text_font@',
                         '|[Font]<>'],
                        '|{Font}'],
                       ['menu@',
                        '|{Menu}'],
                       ['editor@',
                        '|{Editor}'])

    #-------------------------------------------------------------------------
    #  Implementation of the 'label' property:
    #-------------------------------------------------------------------------

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return user_name_for(self.name)

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed('label', old, label)

    #-------------------------------------------------------------------------
    #  Gets the value of the column for a specified object:
    #-------------------------------------------------------------------------

    def get_raw_value(self, object):
        """ Gets the unformatted value of the column for a specified object.
        """
        try:
            return xgetattr(self.get_object(object), self.name)
        except Exception as e:
            if TRAITS_DEBUG:
                from traits.api import raise_to_debug
                raise_to_debug()
            return None

    def get_value(self, object):
        """ Gets the formatted value of the column for a specified object.
        """
        try:
            if self.format_func is not None:
                return self.format_func(self.get_raw_value(object))

            return self.format % (self.get_raw_value(object), )
        except:
            logger.exception('Error occurred trying to format a %s value' %
                             self.__class__.__name__)
            return 'Format!'

    #-------------------------------------------------------------------------
    #  Returns the drag value for the column:
    #-------------------------------------------------------------------------

    def get_drag_value(self, object):
        """Returns the drag value for the column.
        """
        return self.get_raw_value(object)

    #-------------------------------------------------------------------------
    #  Sets the value of the column for a specified object:
    #-------------------------------------------------------------------------

    def set_value(self, object, value):
        """ Sets the value of the column for a specified object.
        """
        target, name = self.target_name(object)
        setattr(target, name, value)

    #-------------------------------------------------------------------------
    #  Gets the editor for the column of a specified object:
    #-------------------------------------------------------------------------

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object.
        """
        if self.editor is not None:
            return self.editor

        target, name = self.target_name(object)

        return target.base_trait(name).get_editor()

    #-------------------------------------------------------------------------
    #  Gets the editor style for the column of a specified object:
    #-------------------------------------------------------------------------

    def get_style(self, object):
        """ Gets the editor style for the column of a specified object.
        """
        return self.style

    #-------------------------------------------------------------------------
    #  Function that gets the value to sort by for a column
    #-------------------------------------------------------------------------

    def key(self, object):
        """ Returns the value to use for sorting.
        """
        return self.get_raw_value(object)

    #-------------------------------------------------------------------------
    #  Returns whether a specified value is valid for dropping on the column
    #  for a specified object:
    #-------------------------------------------------------------------------

    def is_droppable(self, object, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object.
        """
        if self.droppable:
            try:
                target, name = self.target_name(object)
                target.base_trait(name).validate(target, name, value)
                return True
            except:
                pass

        return False

    #-------------------------------------------------------------------------
    #  Returns the target object and name for the column:
    #-------------------------------------------------------------------------

    def target_name(self, object):
        """ Returns the target object and name for the column.
        """
        object = self.get_object(object)
        name = self.name
        col = name.rfind('.')
        if col < 0:
            return (object, name)

        return (xgetattr(object, name[:col]), name[col + 1:])

#-------------------------------------------------------------------------
#  'ExpressionColumn' class:
#-------------------------------------------------------------------------


class ExpressionColumn(ObjectColumn):
    """ A column for displaying computed values.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The Python expression used to return the value of the column:
    expression = Expression

    # Is this column editable?
    editable = Constant(False)

    # The globals dictionary that should be passed to the expression
    # evaluation:
    globals = Any({})

    #-------------------------------------------------------------------------
    #  Gets the value of the column for a specified object:
    #-------------------------------------------------------------------------

    def get_raw_value(self, object):
        """ Gets the unformatted value of the column for a specified object.
        """
        try:
            return eval(self.expression_, self.globals, {'object': object})
        except Exception:
            logger.exception('Error evaluating table column expression: %s' %
                             self.expression)
            return None

#-------------------------------------------------------------------------
#  'NumericColumn' class:
#-------------------------------------------------------------------------


class NumericColumn(ObjectColumn):
    """ A column for editing Numeric arrays.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Column label to use for this column
    label = Property

    # Text color this column when selected
    selected_text_color = Color('black')

    # Text font for this column when selected
    selected_text_font = Font

    # Cell background color for this column when selected
    selected_cell_color = Color(0xD8FFD8)

    # Formatting string for the cell value
    format = Str('%s')

    # Horizontal alignment of text in the column; this value overrides the
    # default.
    horizontal_alignment = 'center'

    #-------------------------------------------------------------------------
    #  Implementation of the 'label' property:
    #-------------------------------------------------------------------------

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return self.name

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed('label', old, label)

    #-------------------------------------------------------------------------
    #  Gets the type of data for the column for a specified object row:
    #-------------------------------------------------------------------------

    def get_type(self, object):
        """ Gets the type of data for the column for a specified object row.
        """
        return self.type

    #-------------------------------------------------------------------------
    #  Returns the text color for the column for a specified object row:
    #-------------------------------------------------------------------------

    def get_text_color(self, object):
        """ Returns the text color for the column for a specified object row.
        """
        if self._is_selected(object):
            return self.selected_text_color_
        return self.text_color_

    #-------------------------------------------------------------------------
    #  Returns the text font for the column for a specified object row:
    #-------------------------------------------------------------------------

    def get_text_font(self, object):
        """ Returns the text font for the column for a specified object row.
        """
        if self._is_selected(object):
            return self.selected_text_font
        return self.text_font

    #-------------------------------------------------------------------------
    #  Returns the cell background color for the column for a specified object
    #  row:
    #-------------------------------------------------------------------------

    def get_cell_color(self, object):
        """ Returns the cell background color for the column for a specified
            object row.
        """
        if self.is_editable(object):
            if self._is_selected(object):
                return self.selected_cell_color_
            return self.cell_color_
        return self.read_only_cell_color_

    #-------------------------------------------------------------------------
    #  Returns the horizontal alignment for the column for a specified object
    #  row:
    #-------------------------------------------------------------------------

    def get_horizontal_alignment(self, object):
        """ Returns the horizontal alignment for the column for a specified
            object row.
        """
        return self.horizontal_alignment

    #-------------------------------------------------------------------------
    #  Returns the vertical alignment for the column for a specified object row:
    #-------------------------------------------------------------------------

    def get_vertical_alignment(self, object):
        """ Returns the vertical alignment for the column for a specified
            object row.
        """
        return self.vertical_alignment

    #-------------------------------------------------------------------------
    #  Returns whether the column is editable for a specified object row:
    #-------------------------------------------------------------------------

    def is_editable(self, object):
        """ Returns whether the column is editable for a specified object row.
        """
        return self.editable

    #-------------------------------------------------------------------------
    #  Returns whether a specified value is valid for dropping on the column
    #  for a specified object row:
    #-------------------------------------------------------------------------

    def is_droppable(self, object, row, value):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object row.
        """
        return self.droppable

    #-------------------------------------------------------------------------
    #  Returns the context menu to display when the user right-clicks on the
    #  column for a specified object row:
    #-------------------------------------------------------------------------

    def get_menu(self, object, row):
        """ Returns the context menu to display when the user right-clicks on
            the column for a specified object row.
        """
        return self.menu

    #-------------------------------------------------------------------------
    #  Gets the value of the column for a specified object row:
    #-------------------------------------------------------------------------

    def get_value(self, object):
        """ Gets the value of the column for a specified object row.
        """
        try:
            value = getattr(object, self.name)
            try:
                return self.format % (value, )
            except:
                return 'Format!'
        except:
            return 'Undefined!'

    #-------------------------------------------------------------------------
    #  Sets the value of the column for a specified object row:
    #-------------------------------------------------------------------------

    def set_value(self, object, row, value):
        """ Sets the value of the column for a specified object row.
        """
        column = self.get_data_column(object)
        column[row] = type(column[row])(value)

    #-------------------------------------------------------------------------
    #  Gets the editor for the column of a specified object row:
    #-------------------------------------------------------------------------

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object row.
        """
        return super(NumericColumn, self).get_editor(object)

    #-------------------------------------------------------------------------
    #  Gets the entire contents of the specified object column:
    #-------------------------------------------------------------------------

    def get_data_column(self, object):
        """ Gets the entire contents of the specified object column.
        """
        return getattr(object, self.name)

    #-------------------------------------------------------------------------
    #  Returns whether a specified object row is selected or not:
    #-------------------------------------------------------------------------

    def _is_selected(self, object):
        """ Returns whether a specified object row is selected.
        """
        if hasattr(object, 'model_selection') \
                and object.model_selection is not None:
            return True
        return False

#-------------------------------------------------------------------------
#  'ListColumn' class:
#-------------------------------------------------------------------------


class ListColumn(TableColumn):
    """ A column for editing lists.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    #Label to use for this column
    label = Property

    # Index of the list element associated with this column
    index = Int

    # Is this column editable? This value overrides the base class default.
    editable = False

    #-------------------------------------------------------------------------
    #  Trait view definitions:
    #-------------------------------------------------------------------------

    traits_view = View([['index', 'label', 'type', '|[Column Information]'],
                        ['text_color@', 'cell_color@', '|[UI Colors]']])

    #-------------------------------------------------------------------------
    #  Implementation of the 'label' property:
    #-------------------------------------------------------------------------

    def _get_label(self):
        """ Gets the label of the column.
        """
        if self._label is not None:
            return self._label
        return 'Column %d' % (self.index + 1)

    def _set_label(self, label):
        old, self._label = self._label, label
        if old != label:
            self.trait_property_changed('label', old, label)

    #-------------------------------------------------------------------------
    #  Gets the value of the column for a specified object:
    #-------------------------------------------------------------------------

    def get_value(self, object):
        """ Gets the value of the column for a specified object.
        """
        return unicode(object[self.index])

    #-------------------------------------------------------------------------
    #  Sets the value of the column for a specified object:
    #-------------------------------------------------------------------------

    def set_value(self, object, value):
        """ Sets the value of the column for a specified object.
        """
        object[self.index] = value

    #-------------------------------------------------------------------------
    #  Gets the editor for the column of a specified object:
    #-------------------------------------------------------------------------

    def get_editor(self, object):
        """ Gets the editor for the column of a specified object.
        """
        return None

    #-------------------------------------------------------------------------
    #  Function that gets the value to sort by for a column
    #-------------------------------------------------------------------------

    def key(self, object):
        """ Returns the value to use for sorting.
        """
        return object[self.index]



# peppy Copyright (c) 2006-2010 Rob McMullen
# Licenced under the GPLv2; see http://peppy.flipturn.org for more info
"""Clipboard utilities for wx

On non-X11 systems, there's only one clipboard.  However, on X11 there are
two: the primary selection and the clipboard.  Due to a current wxWidgets
limitation (due to be resolved in 3.0), only one of those clipboards is
available at a time, and stuffing data into one cancels out the other.
    
On non-X11 systems, there's only one real clipboard and there isn't a
system equivalent to the primary selection.  When on a non-X11 platform,
these routines attempt to mimic the primary selection middle mouse paste.
Unfortunately the primary selection will only be application local if we
aren't on an X11 system.
"""

import wx


# Global boolean use_x11_primary_selection describes whether or not the
# platform supports true X11 style primary selection.
use_x11_primary_selection = None


def setAllowX11PrimarySelection(default=True):
    global use_x11_primary_selection
    try:
        version = wx.version().split('.')
        version = int(version[0]) * 100 + int(version[1]) * 10 + int(version[2])
        if wx.Platform == "__WXGTK__" and version > 287:
            use_x11_primary_selection = default
        else:
            use_x11_primary_selection = False
    except:
        use_x11_primary_selection = False


# Initialize the value for use_x11_primary_selection
if use_x11_primary_selection is None:
    setAllowX11PrimarySelection()

# Global storage non_x11_primary_selection is only used on a non-x11 platform,
# and simulates the primary selection capability of the X11 platform.
# Assuming the STC handles calls SetClipboardText in response to the end of
# a selection event (i.e.  in response to a left mouse up event), the middle
# mouse usage can be simulated on a non-X11 platform.
non_x11_primary_selection = None


def GetClipboardText(primary_selection=False):
    """Returns the current clipboard value.
    
    On non-X11 systems, there's only one clipboard.  However, on X11 there are
    two: the primary selection and the clipboard.  Due to a current wxWidgets
    limitation, only one of those clipboards is available at a time, and
    stuffing data into one cancels out the other.  The calling method will
    have to check the return value to see if data from the requested clipboard
    type is valid.
    
    @param primary_selection: True for the X11 primary selection, False (the
    default) for the normal clipboard.
    """
    global use_x11_primary_selection
    global non_x11_primary_selection

    success = False
    if primary_selection:
        if use_x11_primary_selection:
            wx.TheClipboard.UsePrimarySelection(primary_selection)
        else:
            #dprint(non_x11_primary_selection)
            return non_x11_primary_selection
    do = wx.TextDataObject()
    if wx.TheClipboard.Open():
        success = wx.TheClipboard.GetData(do)
        wx.TheClipboard.Close()

    if success:
        return do.GetText()
    return None


def SetClipboardText(txt, primary_selection=False):
    """Sets the current clipboard value to the given text.
    
    wxWidgets knows about two types of clipboards: the primary selection and
    the clipboard.  On non-X11 systems, there's only one real clipboard and
    there isn't a system equivalent to the primary selection
    
    This method, however, simulates the primary selection when using a non-X11
    system.  It keeps a separate copy of the text so that a middle mouse paste
    can work, although it is only application local and middle paste to some
    other Windows application won't work.
    
    When using a real X11 system, there is a current wxWidgets limitation: only
    one of those clipboards is available at a time.  Stuffing data into one
    cancels out the other.
    
    @param txt: the text string to store in the clipboard
    
    @param primary_selection: True for the X11 primary selection, False (the
    default) for the normal clipboard.
    """
    global use_x11_primary_selection
    global non_x11_primary_selection

    if primary_selection:
        if use_x11_primary_selection:
            wx.TheClipboard.UsePrimarySelection(primary_selection)
        else:
            non_x11_primary_selection = txt
            #dprint(non_x11_primary_selection)
            return 1
    do = wx.TextDataObject()
    do.SetText(txt)
    if wx.TheClipboard.Open():
        wx.TheClipboard.SetData(do)
        wx.TheClipboard.Close()
        return 1
    else:
        eprint("Can't open clipboard!")
    return 0


__all__ = ['GetClipboardText', 'SetClipboardText']



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various instance editors for the wxPython user interface
toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import HasTraits, Property

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.instance_editor file.
from traitsui.editors.instance_editor \
    import ToolkitEditorFactory

from traitsui.ui_traits \
    import AView

from traitsui.helper \
    import user_name_for

from traitsui.handler \
    import Handler

from traitsui.instance_choice \
    import InstanceChoiceItem

from editor \
    import Editor

from constants \
    import DropColor, is_wx26

from helper \
    import TraitsUIPanel, position_window

from pyface.wx.drag_and_drop \
    import PythonDropTarget

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

OrientationMap = {
    'default': None,
    'horizontal': wx.HORIZONTAL,
    'vertical': wx.VERTICAL
}

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(Editor):
    """ Custom style of editor for instances. If selection among instances is
    allowed, the editor displays a combo box listing instances that can be
    selected. If the current instance is editable, the editor displays a panel
    containing trait editors for all the instance's traits.
    """

    # Background color when an item can be dropped on the editor:
    ok_color = DropColor

    # The orientation of the instance editor relative to the instance selector:
    orientation = wx.VERTICAL

    # Class constant:
    extra = 0

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # List of InstanceChoiceItem objects used by the editor
    items = Property

    # The maximum extra padding that should be allowed around the editor:
    # (Override of the Editor base class trait)
    border_size = 0

    # The view to use for displaying the instance:
    view = AView

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if factory.name != '':
            self._object, self._name, self._value = \
                self.parse_extended_name(factory.name)

        # Create a panel to hold the object trait's view:
        if factory.editable:
            self.control = self._panel = parent = TraitsUIPanel(parent, -1)

        # Build the instance selector if needed:
        selectable = factory.selectable
        droppable = factory.droppable
        items = self.items
        for item in items:
            droppable |= item.is_droppable()
            selectable |= item.is_selectable()

        if selectable:
            self._object_cache = {}
            item = self.item_for(self.value)
            if item is not None:
                self._object_cache[id(item)] = self.value

            self._choice = choice = wx.Choice(parent, -1, wx.Point(0, 0),
                                              wx.Size(-1, -1), [])
            wx.EVT_CHOICE(choice, choice.GetId(), self.update_object)
            if droppable:
                self._choice.SetBackgroundColour(self.ok_color)

            self.set_tooltip(self._choice)

            if factory.name != '':
                self._object.on_trait_change(self.rebuild_items,
                                             self._name, dispatch='ui')
                self._object.on_trait_change(
                    self.rebuild_items, self._name + '_items', dispatch='ui')

            factory.on_trait_change(self.rebuild_items, 'values',
                                    dispatch='ui')
            factory.on_trait_change(self.rebuild_items, 'values_items',
                                    dispatch='ui')

            self.rebuild_items()

        elif droppable:
            self._choice = wx.TextCtrl(parent, -1, '',
                                       style=wx.TE_READONLY)
            self._choice.SetBackgroundColour(self.ok_color)
            self.set_tooltip(self._choice)

        if droppable:
            self._choice.SetDropTarget(PythonDropTarget(self))

        orientation = OrientationMap[factory.orientation]
        if orientation is None:
            orientation = self.orientation

        if (selectable or droppable) and factory.editable:
            sizer = wx.BoxSizer(orientation)
            sizer.Add(self._choice, self.extra, wx.EXPAND)
            if orientation == wx.VERTICAL:
                sizer.Add(
                    wx.StaticLine(parent, -1, style=wx.LI_HORIZONTAL), 0,
                    wx.EXPAND | wx.TOP | wx.BOTTOM, 5)
            self.create_editor(parent, sizer)
            parent.SetSizer(sizer)
        elif self.control is None:
            if self._choice is None:
                self._choice = choice = wx.Choice(parent, -1, wx.Point(0, 0),
                                                  wx.Size(-1, -1), [])
                wx.EVT_CHOICE(choice, choice.GetId(), self.update_object)
            self.control = self._choice
        else:
            sizer = wx.BoxSizer(orientation)
            self.create_editor(parent, sizer)
            parent.SetSizer(sizer)

        # Synchronize the 'view' to use:
        # fixme: A normal assignment can cause a crash (for unknown reasons) in
        # some cases, so we make sure that no notifications are generated:
        self.trait_setq(view=factory.view)
        self.sync_value(factory.view_name, 'view', 'from')

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        # Make sure we aren't hanging on to any object refs:
        self._object_cache = None

        if self._ui is not None:
            self._ui.dispose()

        choice = self._choice
        if choice is not None:
            if isinstance(choice, wx.Choice):
                wx.EVT_CHOICE(choice, choice.GetId(), None)

            if self._object is not None:
                self._object.on_trait_change(self.rebuild_items,
                                             self._name, remove=True)
                self._object.on_trait_change(
                    self.rebuild_items, self._name + '_items', remove=True)

            self.factory.on_trait_change(self.rebuild_items, 'values',
                                         remove=True)
            self.factory.on_trait_change(self.rebuild_items,
                                         'values_items', remove=True)

        super(CustomEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Creates the editor control:
    #-------------------------------------------------------------------------

    def create_editor(self, parent, sizer):
        """ Creates the editor control.
        """
        self._panel = TraitsUIPanel(parent, -1)
        sizer.Add(self._panel, 1, wx.EXPAND)

    #-------------------------------------------------------------------------
    #  Gets the current list of InstanceChoiceItem items:
    #-------------------------------------------------------------------------

    def _get_items(self):
        """ Gets the current list of InstanceChoiceItem items.
        """
        if self._items is not None:
            return self._items

        factory = self.factory
        if self._value is not None:
            values = self._value() + factory.values
        else:
            values = factory.values

        items = []
        adapter = factory.adapter
        for value in values:
            if not isinstance(value, InstanceChoiceItem):
                value = adapter(object=value)
            items.append(value)

        self._items = items

        return items

    #-------------------------------------------------------------------------
    #  Rebuilds the object selector list:
    #-------------------------------------------------------------------------

    def rebuild_items(self):
        """ Rebuilds the object selector list.
        """
        # Clear the current cached values:
        self._items = None

        # Rebuild the contents of the selector list:
        name = None
        value = self.value
        choice = self._choice
        choice.Clear()
        for item in self.items:
            if item.is_selectable():
                item_name = item.get_name()
                choice.Append(item_name)
                if item.is_compatible(value):
                    name = item_name

        # Reselect the current item if possible:
        if name is not None:
            choice.SetStringSelection(name)
        else:
            # Otherwise, current value is no longer valid, try to discard it:
            try:
                self.value = None
            except:
                pass

    #-------------------------------------------------------------------------
    #  Returns the InstanceChoiceItem for a specified object:
    #-------------------------------------------------------------------------

    def item_for(self, object):
        """ Returns the InstanceChoiceItem for a specified object.
        """
        for item in self.items:
            if item.is_compatible(object):
                return item

        return None

    #-------------------------------------------------------------------------
    #  Returns the view to use for a specified object:
    #-------------------------------------------------------------------------

    def view_for(self, object, item):
        """ Returns the view to use for a specified object.
        """
        view = ''
        if item is not None:
            view = item.get_view()

        if view == '':
            view = self.view

        return self.ui.handler.trait_view_for(self.ui.info, view, object,
                                              self.object_name, self.name)

    #-------------------------------------------------------------------------
    #  Handles the user selecting a new value from the combo box:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user selecting a new value from the combo box.
        """
        name = event.GetString()
        for item in self.items:
            if name == item.get_name():
                id_item = id(item)
                object = self._object_cache.get(id_item)
                if object is None:
                    object = item.get_object()
                    if (not self.factory.editable) and item.is_factory:
                        view = self.view_for(object, self.item_for(object))
                        view.ui(object, self.control, 'modal')

                    if self.factory.cachable:
                        self._object_cache[id_item] = object

                self.value = object
                self.resynch_editor()
                break

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        # Attach the current object value to the control (for use by
        # DockWindowFeature):

        # fixme: This code is somewhat fragile since it assumes that if a
        # DockControl is involved, the parent of this editor will be the
        # control being managed by the DockControl.
        parent = self.control.GetParent()
        parent._object = self.value
        dock_control = getattr(parent, '_dock_control', None)
        if dock_control is not None:
            dock_control.reset_tab()

        # Synchronize the editor contents:
        self.resynch_editor()

        # Update the selector (if any):
        choice = self._choice
        item = self.item_for(self.value)
        if (choice is not None) and (item is not None):
            name = item.get_name(self.value)
            if self._object_cache is not None:
                if choice.FindString(name) < 0:
                    choice.Append(name)
                choice.SetStringSelection(name)
            else:
                choice.SetValue(name)

    #-------------------------------------------------------------------------
    #  Resynchronizes the contents of the editor when the object trait changes
    #  external to the editor:
    #-------------------------------------------------------------------------

    def resynch_editor(self):
        """ Resynchronizes the contents of the editor when the object trait
        changes externally to the editor.
        """
        panel = self._panel
        if panel is not None:
            # Compute/update the maximum size the panel has ever been:
            dx, dy = panel.GetSizeTuple()
            mdx = mdy = 0
            if self._panel_size is not None:
                mdx, mdy = self._panel_size
            self._panel_size = size = wx.Size(max(mdx, dx), max(mdy, dy))

            # Dispose of the previous contents of the panel:
            panel.SetSizer(None)
            if self._ui is not None:
                self._ui.dispose()
                self._ui = None
            else:
                for child in panel.GetChildren():
                    while child.GetEventHandler() is not child:
                        child.PopEventHandler(True)
                    child.Destroy()

            # Create the new content for the panel:
            sizer = wx.BoxSizer(wx.VERTICAL)
            stretch = 0
            value = self.value
            if not isinstance(value, HasTraits):
                str_value = ''
                if value is not None:
                    str_value = self.str_value
                control = wx.StaticText(panel, -1, str_value)
            else:
                view = self.view_for(value, self.item_for(value))
                context = value.trait_context()
                handler = None
                if isinstance(value, Handler):
                    handler = value
                context.setdefault('context', self.object)
                context.setdefault('context_handler', self.ui.handler)
                self._ui = ui = view.ui(context, panel, 'subpanel',
                                        value.trait_view_elements(), handler,
                                        self.factory.id)
                control = ui.control
                self.scrollable = ui._scrollable
                ui.parent = self.ui

                if view.resizable or view.scrollable or ui._scrollable:
                    stretch = 1

            # Make sure the panel and its contents are correctly sized (This
            # code is complicated by the various layout bugs present in wx.
            # Tamper with it at your own risk!):
            control.Freeze()
            if stretch and (size != (20, 20)):
                control.SetSize(size)
                panel.SetSize(size)
            else:
                panel.SetSize(control.GetSize())
            sizer.Add(control, stretch, wx.EXPAND)
            panel.SetSizer(sizer)
            control.Thaw()
            self.control.Layout()
            parent = self.control.GetParent()
            parent.Layout()

            # It is possible that this instance editor is embedded at some level
            # in a ScrolledWindow. If so, we need to inform the window that the
            # size of the editor's contents have (potentially) changed:
            # NB: There is a typo in the wxPython 2.6 code that prevents the
            # 'SendSizeEvent' from working correctly, so we just skip it.
            if not is_wx26:
                while ((parent is not None) and
                       (not isinstance(parent, wx.ScrolledWindow))):
                    parent = parent.GetParent()

                if parent is not None:
                    parent.SendSizeEvent()

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        pass

    #-------------------------------------------------------------------------
    #  Returns the editor's control for indicating error status:
    #-------------------------------------------------------------------------

    def get_error_control(self):
        """ Returns the editor's control for indicating error status.
        """
        return (self._choice or self.control)

    #-- UI preference save/restore interface ---------------------------------

    #-------------------------------------------------------------------------
    #  Restores any saved user preference information associated with the
    #  editor:
    #-------------------------------------------------------------------------

    def restore_prefs(self, prefs):
        """ Restores any saved user preference information associated with the
            editor.
        """
        ui = self._ui
        if (ui is not None) and (prefs.get('id') == ui.id):
            ui.set_prefs(prefs.get('prefs'))

    #-------------------------------------------------------------------------
    #  Returns any user preference information associated with the editor:
    #-------------------------------------------------------------------------

    def save_prefs(self):
        """ Returns any user preference information associated with the editor.
        """
        ui = self._ui
        if (ui is not None) and (ui.id != ''):
            return {'id': ui.id,
                    'prefs': ui.get_prefs()}

        return None

    #-- Drag and drop event handlers -----------------------------------------

    #-------------------------------------------------------------------------
    #  Handles a Python object being dropped on the control:
    #-------------------------------------------------------------------------

    def wx_dropped_on(self, x, y, data, drag_result):
        """ Handles a Python object being dropped on the tree.
        """
        for item in self.items:
            if item.is_droppable() and item.is_compatible(data):
                if self._object_cache is not None:
                    self.rebuild_items()
                self.value = data
                return drag_result

        return wx.DragNone

    #-------------------------------------------------------------------------
    #  Handles a Python object being dragged over the control:
    #-------------------------------------------------------------------------

    def wx_drag_over(self, x, y, data, drag_result):
        """ Handles a Python object being dragged over the tree.
        """
        for item in self.items:
            if item.is_droppable() and item.is_compatible(data):
                return drag_result

        return wx.DragNone

    #-- Traits event handlers ------------------------------------------------

    def _view_changed(self, view):
        self.resynch_editor()

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(CustomEditor):
    """ Simple style of editor for instances, which displays a button. Clicking
    the button displays a dialog box in which the instance can be edited.
    """

    # Class constants:
    orientation = wx.HORIZONTAL
    extra = 2

    #-------------------------------------------------------------------------
    #  Creates the editor control:
    #-------------------------------------------------------------------------

    def create_editor(self, parent, sizer):
        """ Creates the editor control (a button).
        """
        self._button = button = wx.Button(parent, -1, '')
        sizer.Add(button, 1, wx.EXPAND | wx.LEFT, 5)
        wx.EVT_BUTTON(button, button.GetId(), self.edit_instance)

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        button = self._button
        if button is not None:
            wx.EVT_BUTTON(button, button.GetId(), None)

        super(SimpleEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Edit the contents of the object trait when the user clicks the button:
    #-------------------------------------------------------------------------

    def edit_instance(self, event):
        """ Edit the contents of the object trait when the user clicks the
            button.
        """
        # Create the user interface:
        factory = self.factory
        view = self.ui.handler.trait_view_for(self.ui.info, factory.view,
                                              self.value, self.object_name,
                                              self.name)
        ui = self.value.edit_traits(view, self.control, factory.kind,
                                    id=factory.id)

        # Check to see if the view was 'modal', in which case it will already
        # have been closed (i.e. is None) by the time we get control back:
        if ui.control is not None:
            # Position the window on the display:
            position_window(ui.control)

            # Chain our undo history to the new user interface if it does not
            # have its own:
            if ui.history is None:
                ui.history = self.ui.history

    #-------------------------------------------------------------------------
    #  Resynchronizes the contents of the editor when the object trait changes
    #  external to the editor:
    #-------------------------------------------------------------------------

    def resynch_editor(self):
        """ Resynchronizes the contents of the editor when the object trait
            changes externally to the editor.
        """
        button = self._button
        if button is not None:
            label = self.factory.label
            if label == '':
                label = user_name_for(self.name)
            button.SetLabel(label)
            button.Enable(isinstance(self.value, HasTraits))

### EOF #######################################################################



# Major package imports.
import wx

import logging
log = logging.getLogger(__name__)


class AboutDialog(object):
    """ The toolkit specific implementation of an AboutDialog.  See the
    IAboutDialog interface for the API documentation.
    """

    ###########################################################################
    # Protected 'IDialog' interface.
    ###########################################################################

    def __init__(self, parent, about):
        info = wx.AboutDialogInfo()

        # Load the image to be displayed in the about box.
        #image = self.about_image.create_image()
        icon = wx.EmptyIcon()
        try:
            icon.CopyFromBitmap(about.about_image.create_bitmap())
            info.SetIcon(icon)
        except:
            log.error("AboutDialog: bad icon file: %s" % about.about_image.absolute_path)

        info.SetName(about.about_title)
        info.SetVersion(about.about_version)
        info.SetDescription(about.about_description)
        info.SetWebSite(about.about_website)

        dialog = wx.AboutBox(info)



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------

from adapter_base import AdapterBase
from i_policy_manager import IPolicyManager
from i_user import IUser
from i_user_manager import IUserManager
from package_globals import get_permissions_manager, set_permissions_manager
from permission import ManagePolicyPermission, ManageUsersPermission, Permission
from permissions_manager import PermissionsManager
from secure_proxy import SecureHandler, SecureProxy



#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt

from __future__ import absolute_import

from traits.api import (Bool, Dict, Either, Enum, Font, Instance, List,
                        Property, Str)

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.editors.tabular_editor import TabularEditor
from traitsui.item import Item
from traitsui.tabular_adapter import TabularAdapter
from traitsui.toolkit import toolkit_object
from traitsui.ui_editor import UIEditor
from traitsui.view import View


class DataFrameAdapter(TabularAdapter):
    """ Generic tabular adapter for data frames
    """

    #: The text to use for a generic entry.
    text = Property

    #: The alignment for each cell
    alignment = Property(Enum('left', 'center', 'right'))

    #: The text to use for a row index.
    index_text = Property

    #: The alignment to use for a row index.
    index_alignment = Property

    #: The font to use for each column
    font = Property

    #: The format to use for each column
    format = Property

    #: The format for each element, or a mapping column ID to format.
    _formats = Either(Str, Dict, default='%s')

    #: The font for each element, or a mapping column ID to font.
    _fonts = Either(Font, Dict, default='Courier 10')

    def _get_index_alignment(self):
        import numpy as np

        index = getattr(self.object, self.name).index
        if np.issubdtype(index.dtype, np.number):
            return 'right'
        else:
            return 'left'

    def _get_alignment(self):
        import numpy as np

        column = self.item[self.column_id]
        if np.issubdtype(column.dtype, np.number):
            return 'right'
        else:
            return 'left'

    def _get_font(self):
        if isinstance(self._fonts, toolkit_object('font_trait:TraitsFont')):
            return self._fonts
        else:
            return self._fonts.get(self.column_id, 'Courier 10')

    def _get_format(self):
        if isinstance(self._formats, basestring):
            return self._formats
        else:
            return self._formats.get(self.column_id, '%s')

    def _get_content(self):
        return self.item[self.column_id].iloc[0]

    def _get_text(self):
        format = self.get_format(self.object, self.name, self.row, self.column)
        return format % self.get_content(self.object, self.name, self.row,
                                         self.column)

    def _set_text(self, value):
        column = self.item[self.column_id]
        dtype = column.dtype
        value = dtype.type(value)
        column.iloc[0] = value

    def _get_index_text(self):
        return str(self.item.index[0])

    def _set_index_text(self, value):
        index = getattr(self.object, self.name).index
        dtype = index.dtype
        value = dtype.type(value)
        index.values[self.row] = value

    #---- Adapter methods that are not sensitive to item type ----------------

    def get_item(self, object, trait, row):
        """ Override the base implementation to work with DataFrames

        This returns a dataframe with one row, rather than a series, since
        using a dataframe preserves dtypes.

        """
        return getattr(object, trait).iloc[row:row + 1]

    def delete(self, object, trait, row):
        """ Override the base implementation to work with DataFrames

        Unavoidably does a copy of the data, setting the trait with the new
        value.
        """
        import pandas as pd

        df = getattr(object, trait)
        if 0 < row < len(df) - 1:
            new_df = pd.concat([df.iloc[:row, :], df.iloc[row + 1:, :]])
        elif row == 0:
            new_df = df.iloc[row + 1:, :]
        else:
            new_df = df.iloc[:row, :]
        setattr(object, trait, new_df)

    def insert(self, object, trait, row, value):
        """ Override the base implementation to work with DataFrames

        Unavoidably does a copy of the data, setting the trait with the new
        value.
        """
        import pandas as pd

        df = getattr(object, trait)
        if 0 < row < len(df) - 1:
            new_df = pd.concat([df.iloc[:row, :], value, df.iloc[row:, :]])
        elif row == 0:
            new_df = pd.concat([value, df])
        else:
            new_df = pd.concat([df, value])
        setattr(object, trait, new_df)


class _DataFrameEditor(UIEditor):
    """ TraitsUI-based editor implementation for data frames """

    # Indicate that the editor is scrollable/resizable:
    scrollable = True

    # Should column titles be displayed:
    show_titles = Bool(True)

    # The tabular adapter being used for the editor view:
    adapter = Instance(DataFrameAdapter)

    #-- Private Methods ------------------------------------------------------

    def _target_name(self, name):
        if name:
            return 'object.object.' + name
        else:
            return ''

    def _data_frame_view(self):
        """ Return the view used by the editor.
        """

        return View(
            Item(
                self._target_name(self.name),
                id='tabular_editor',
                show_label=False,
                editor=TabularEditor(
                    show_titles=self.factory.show_titles,
                    editable=self.factory.editable,
                    adapter=self.adapter,
                    selected=self._target_name(self.factory.selected),
                    selected_row=self._target_name(self.factory.selected_row),
                    selectable=self.factory.selectable,
                    activated=self._target_name(self.factory.selected),
                    activated_row=self._target_name(self.factory.activated_row),  # noqa
                    clicked=self._target_name(self.factory.clicked),
                    dclicked=self._target_name(self.factory.dclicked),
                    right_clicked=self._target_name(self.factory.right_clicked),  # noqa
                    right_dclicked=self._target_name(self.factory.right_dclicked),  # noqa
                    column_clicked=self._target_name(self.factory.column_clicked),  # noqa
                    column_right_clicked=self._target_name(self.factory.column_right_clicked),  # noqa
                    operations=self.factory.operations,
                    update=self._target_name(self.factory.update),
                    refresh=self._target_name(self.factory.refresh),
                )
            ),
            id='data_frame_editor',
            resizable=True
        )

    def init_ui(self, parent):
        """ Creates the Traits UI for displaying the array.
        """
        factory = self.factory
        if (factory.columns != []):
            columns = []
            for column in factory.columns:
                if isinstance(column, basestring):
                    title = column
                    column_id = column
                else:
                    title, column_id = column
                if column_id not in self.value.columns:
                    continue
                columns.append((title, column_id))
        else:
            columns = [(column_id, column_id)
                       for column_id in self.value.columns]

        if factory.show_index:
            index_name = self.value.index.name
            if index_name is None:
                index_name = ''
            columns.insert(0, (index_name, 'index'))

        if factory.adapter is not None:
            self.adapter = factory.adapter
            self.adapter._formats=factory.formats
            self.adapter._fonts=factory.fonts
            if not self.adapter.columns:
                self.adapter.columns = columns
        else:
            self.adapter = DataFrameAdapter(
                columns=columns,
                _formats=factory.formats,
                _fonts=factory.fonts
            )

        return self.edit_traits(
            view='_data_frame_view',
            parent=parent,
            kind='subpanel'
        )


class DataFrameEditor(BasicEditorFactory):
    """ Editor factory for basic data frame editor """

    #: The editor implementation class.
    klass = Property

    #: Should an index column be displayed.
    show_index = Bool(True)

    #: Should column headers be displayed.
    show_titles = Bool(True)

    #: Optional list of either column ID or pairs of (column title, column ID).
    columns = List()

    #: The format for each element, or a mapping column ID to format.
    formats = Either(Str, Dict, default='%s')

    #: The font for each element, or a mapping column ID to font.
    fonts = Either(Font, Dict, default='Courier 10')

    # The optional extended name of the trait to synchronize the selection
    # values with:
    selected = Str

    # The optional extended name of the trait to synchronize the selection rows
    # with:
    selected_row = Str

    # Whether or not to allow selection.
    selectable = Bool(True)

    # The optional extended name of the trait to synchronize the activated
    # value with:
    activated = Str

    # The optional extended name of the trait to synchronize the activated
    # value's row with:
    activated_row = Str

    # The optional extended name of the trait to synchronize left click data
    # with. The data is a TabularEditorEvent:
    clicked = Str

    # The optional extended name of the trait to synchronize left double click
    # data with. The data is a TabularEditorEvent:
    dclicked = Str

    # The optional extended name of the trait to synchronize right click data
    # with. The data is a TabularEditorEvent:
    right_clicked = Str

    # The optional extended name of the trait to synchronize right double
    # clicked data with. The data is a TabularEditorEvent:
    right_dclicked = Str

    # The optional extended name of the trait to synchronize column
    # clicked data with. The data is a TabularEditorEvent:
    column_clicked = Str

    # The optional extended name of the trait to synchronize column
    # right clicked data with. The data is a TabularEditorEvent:
    column_right_clicked = Str

    #: Whether or not the entries can be edited.
    editable = Bool(False)

    # What type of operations are allowed on the list:
    operations = List(Enum('delete', 'insert', 'append', 'edit', 'move'),
                      ['delete', 'insert', 'append', 'edit', 'move'])

    # The optional extended name of the trait used to indicate that a complete
    # table update is needed:
    update = Str

    # The optional extended name of the trait used to indicate that the table
    # just needs to be repainted.
    refresh = Str

    # Set to override the default dataframe adapter
    adapter = Instance(DataFrameAdapter)

    def _get_klass(self):
        """ The class used to construct editor objects.
        """
        return toolkit_object('data_frame_editor:_DataFrameEditor')



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Ioannis Tziakos
#  Date:   11 Jan 2012
#
#------------------------------------------------------------------------------

""" Defines the various text editors for the wxPython user interface toolkit.
    The module is mainly a place-folder for TextEditor factories that have
    been augmented to also listen to changes in the items of the list object.

"""

#-------------------------------------------------------------------------
#  Imports:
#------------------------------------------------------------------------------

from .text_editor import SimpleEditor as WXSimpleEditor
from .text_editor import CustomEditor as WXCustomEditor
from .text_editor import ReadonlyEditor as WXReadonlyEditor
from ..editors.csv_list_editor import _prepare_method, _dispose_method


class SimpleEditor(WXSimpleEditor):
    """ Simple Editor style for CSVListEditor. """
    prepare = _prepare_method
    dispose = _dispose_method


class CustomEditor(WXCustomEditor):
    """ Custom Editor style for CSVListEditor. """
    prepare = _prepare_method
    dispose = _dispose_method


class ReadonlyEditor(WXReadonlyEditor):
    """ Readonly Editor style for CSVListEditor. """
    prepare = _prepare_method
    dispose = _dispose_method

TextEditor = SimpleEditor



def collapse_json(text, indent=8):
    """Compacts a string of json data by collapsing whitespace after the specified
    indent level
    
    NOTE: will not produce correct results when indent level is not a multiple
    of the json indent level
    """
    initial = " " * indent
    out = []  # final json output
    sublevel = []  # accumulation list for sublevel entries
    pending = None  # holder for consecutive entries at exact indent level
    for line in text.splitlines():
        if line.startswith(initial):
            if line[indent] == " ":
                # found a line indented further than the indent level, so add
                # it to the sublevel list
                if pending:
                    # the first item in the sublevel will be the pending item
                    # that was the previous line in the json
                    sublevel.append(pending)
                    pending = None
                item = line.strip()
                sublevel.append(item)
                if item.endswith(","):
                    sublevel.append(" ")
            elif sublevel:
                # found a line at the exact indent level *and* we have sublevel
                # items. This means the sublevel items have come to an end
                sublevel.append(line.strip())
                out.append("".join(sublevel))
                sublevel = []
            else:
                # found a line at the exact indent level but no items indented
                # further, so possibly start a new sub-level
                if pending:
                    # if there is already a pending item, it means that
                    # consecutive entries in the json had the exact same
                    # indentation and that last pending item was not the start
                    # of a new sublevel.
                    out.append(pending)
                pending = line.rstrip()
        else:
            if pending:
                # it's possible that an item will be pending but not added to
                # the output yet, so make sure it's not forgotten.
                out.append(pending)
                pending = None
            if sublevel:
                out.append("".join(sublevel))
            out.append(line)
    return "\n".join(out)


def dict_to_list(d):
    """Return a sorted list of lists that represent the dictionary.

    Only strings are allowed as keys in a json dictionary, so this transform
    the dictionary into a list of lists that can be serialized in as small
    amount of text as possible. Tuples would be more efficient in python terms,
    but json can't serialize tuples directly either.
    """
    return sorted([list(i) for i in d.iteritems()])


if __name__ == "__main__":
    import json

    s = {"zero": ["first", {"second": 2, "third": 3, "fourth": 4, "items": [[1,2,3,4], [5,6,7,8], 9, 10, [11, [12, [13, [14, 15]]]]]}]}

    text = json.dumps(s, indent=4)

    for level in range(0, 21, 4):
        processed = collapse_json(text, indent=level)
        print level
        print processed
        rebuilt = json.loads(processed)
        assert rebuilt == s



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the Group class used to represent a group of items used in a
    Traits-based user interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (Bool, Delegate, Float, Instance, List, Property, Range,
                        ReadOnly, Str, TraitError, cached_property)

import traits._py2to3 as _py2to3

from .view_element import ViewSubElement

from .item import Item

from .include import Include

from .ui_traits import SequenceTypes, ATheme, ContainerDelegate, Orientation, Layout

from .dock_window_theme import dock_window_theme, DockWindowTheme

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Delegate trait to the object being "shadowed"
ShadowDelegate = Delegate('shadow')

# Amount of padding to add around item
Padding = Range(0, 15, desc='amount of padding to add around each item')

#-------------------------------------------------------------------------
#  'Group' class:
#-------------------------------------------------------------------------


class Group(ViewSubElement):
    """ Represents a grouping of items in a user interface view.
    """

    #-------------------------------------------------------------------------
    # Trait definitions:
    #-------------------------------------------------------------------------

    # A list of Group, Item, and Include objects in this group.
    content = List(ViewSubElement)

    # A unique identifier for the group.
    id = Str

    # User interface label for the group. How the label is displayed depends
    # on the **show_border** attribute, and on the **layout** attribute of
    # the group's parent group or view.
    label = Str

    style_sheet = Str

    # Default context object for group items.
    object = ContainerDelegate

    # Default editor style of items in the group.
    style = ContainerDelegate

    # Default docking style of items in group.
    dock = ContainerDelegate

    # Default image to display on notebook tabs.
    image = ContainerDelegate

    # The theme to use for a DockWindow:
    dock_theme = Instance(DockWindowTheme, allow_none=False)

    # The theme to use for the group itself:
    group_theme = ATheme

    # The theme to use for items contained in the group:
    item_theme = ContainerDelegate

    # The theme to use for the labels of items contained in the group:
    label_theme = ContainerDelegate

    # Category of elements dragged from view.
    export = ContainerDelegate

    # Spatial orientation of the group's elements. Can be 'vertical' (default)
    # or 'horizontal'.
    orientation = Orientation

    # Layout style of the group, which can be one of the following:
    #
    # * 'normal' (default): Sub-groups are displayed sequentially in a single
    #   panel.
    # * 'flow': Sub-groups are displayed sequentially, and then "wrap" when
    #   they exceed the available space in the **orientation** direction.
    # * 'split': Sub-groups are displayed in a single panel, separated by
    #   "splitter bars", which the user can drag to adjust the amount of space
    #   for each sub-group.
    # * 'tabbed': Each sub-group appears on a separate tab, labeled with the
    #   sub-group's *label* text, if any.
    #
    # This attribute is ignored for groups that contain only items, or contain
    # only one sub-group.
    layout = Layout

    # Should the group be scrollable along the direction of orientation?
    scrollable = Bool(False)

    # The number of columns in the group
    columns = Range(1, 50)

    # Should a border be drawn around group? If set to True, the **label** text
    # is embedded in the border. If set to False, the label appears as a banner
    # above the elements of the group.
    show_border = Bool(False)

    # Should labels be added to items in group? Only items that are directly
    # contained in the group are affected. That is, if the group contains
    # a sub-group, the display of labels in the sub-group is not affected by
    # the attribute on this group.
    show_labels = Bool(True)

    # Should labels be shown to the left of items (True) or the right (False)?
    # Only items that are directly contained in the group are affected. That is,
    # if the group contains a sub-group, the display of labels in the sub-group
    # is not affected by the attribute in this group. If **show_labels** is
    # False, this attribute is irrelevant.
    show_left = Bool(True)

    # Is this group the tab that is initially selected? If True, the group's
    # tab is displayed when the view is opened. If the **layout** of the group's
    # parent is not 'tabbed', this attribute is ignored.
    selected = Bool(False)

    # Should the group use extra space along its parent group's layout
    # orientation?
    springy = Bool(False)

    # Optional help text (for top-level group). This help text appears in the
    # View-level help window (created by the default help handler), for any
    # View that contains *only* this group. Group-level help is ignored for
    # nested groups and multiple top-level groups
    help = Str

    # Pre-condition for including the group in the display. If the expression
    # evaluates to False, the group is not defined in the display. Conditions
    # for **defined_when** are evaluated only once, when the display is first
    # constructed. Use this attribute for conditions based on attributes that
    # vary from object to object, but that do not change over time.
    defined_when = Str

    # Pre-condition for showing the group. If the expression evaluates to False,
    # the group and its items are not visible (and they disappear if they were
    # previously visible). If the value evaluates to True, the group and items
    # become visible. All **visible_when** conditions are checked each time
    # that any trait value is edited in the display. Therefore, you can use
    # **visible_when** conditions to hide or show groups in response to user
    # input.
    visible_when = Str

    # Pre-condition for enabling the group. If the expression evaluates to False,
    # the group is disabled, that is, none of the widgets accept input. All
    # **enabled_when** conditions are checked each time that any trait value
    # is edited in the display. Therefore, you can use **enabled_when**
    # conditions to enable or disable groups in response to user input.
    enabled_when = Str

    # Amount of padding (in pixels) to add around each item in the group. The
    # value must be an integer between 0 and 15. (Unlike the Item class, the
    # Group class does not support negative padding.) The padding for any
    # individual widget is the sum of the padding for its Group, the padding
    # for its Item, and the default spacing determined by the toolkit.
    padding = Padding

    # Requested width of the group (calculated from widths of contents)
    width = Property(Float, depends_on='content')

    # Requested height of the group (calculated from heights of contents)
    height = Property(Float, depends_on='content')

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, *values, **traits):
        """ Initializes the group object.
        """
        super(ViewSubElement, self).__init__(**traits)

        content = self.content

        # Process any embedded Group options first:
        for value in values:
            if (isinstance(value, basestring)) and (value[0:1] in '-|'):
                # Parse Group trait options if specified as a string:
                self._parse(value)

        # Process all of the data passed to the constructor:
        for value in values:
            if isinstance(value, ViewSubElement):
                content.append(value)
            elif type(value) in SequenceTypes:
                # Map (...) or [...] to a Group():
                content.append(Group(*value))
            elif isinstance(value, basestring):
                if value[0:1] in '-|':
                    # We've already parsed Group trait options above:
                    pass
                elif (value[:1] == '<') and (value[-1:] == '>'):
                    # Convert string to an Include value:
                    content.append(Include(value[1:-1].strip()))
                else:
                    # Else let the Item class try to make sense of it:
                    content.append(Item(value))
            else:
                raise TypeError("Unrecognized argument type: %s" % value)

        # Make sure this Group is the container for all its children:
        self.set_container()

    #-- Default Trait Values -------------------------------------------------

    def _dock_theme_default(self):
        return dock_window_theme()

    #-------------------------------------------------------------------------
    #  Gets the label to use for a specified Group in a specified UI:
    #-------------------------------------------------------------------------

    def get_label(self, ui):
        """ Gets the label to use this group.
        """
        if self.label != '':
            return self.label

        return 'Group'

    #-------------------------------------------------------------------------
    #  Returns whether or not the object is replacable by an Include object:
    #-------------------------------------------------------------------------

    def is_includable(self):
        """ Returns a Boolean value indicating whether the object is replacable
        by an Include object.
        """
        return (self.id != '')

    #-------------------------------------------------------------------------
    #  Replaces any items which have an 'id' with an Include object with the
    #  same 'id', and puts the object with the 'id' into the specified
    #  ViewElements object:
    #-------------------------------------------------------------------------

    def replace_include(self, view_elements):
        """ Replaces any items that have an **id** attribute with an Include
        object with the same ID value, and puts the object with the ID
        into the specified ViewElements object.

        Parameters
        ----------
        view_elements : ViewElements object
            A set of Group, Item, and Include objects
        """
        for i, item in enumerate(self.content):
            if item.is_includable():
                id = item.id
                if id in view_elements.content:
                    raise TraitError(
                        "Duplicate definition for view element '%s'" % id)
                self.content[i] = Include(id)
                view_elements.content[id] = item
            item.replace_include(view_elements)

    #-------------------------------------------------------------------------
    #  Returns a ShadowGroup for the Group which recursively resolves all
    #  imbedded Include objects and which replaces all imbedded Group objects
    #  with a corresponding ShadowGroup:
    #-------------------------------------------------------------------------

    def get_shadow(self, ui):
        """ Returns a ShadowGroup object for the current Group object, which
        recursively resolves all embedded Include objects and which replaces
        each embedded Group object with a corresponding ShadowGroup.
        """
        content = []
        groups = 0
        level = ui.push_level()
        for value in self.content:
            # Recursively replace Include objects:
            while isinstance(value, Include):
                value = ui.find(value)

            # Convert Group objects to ShadowGroup objects, but include Item
            # objects as is (ignore any 'None' values caused by a failed
            # Include):
            if isinstance(value, Group):
                if self._defined_when(ui, value):
                    content.append(value.get_shadow(ui))
                    groups += 1
            elif isinstance(value, Item):
                if self._defined_when(ui, value):
                    content.append(value)

            ui.pop_level(level)

        # Return the ShadowGroup:
        return ShadowGroup(shadow=self, content=content, groups=groups)

    #-------------------------------------------------------------------------
    #  Sets the correct container for the content:
    #-------------------------------------------------------------------------

    def set_container(self):
        """ Sets the correct container for the content.
        """
        for item in self.content:
            item.container = self

    #-------------------------------------------------------------------------
    #  Returns whether the object should be defined in the user interface:
    #-------------------------------------------------------------------------

    def _defined_when(self, ui, value):
        """ Should the object be defined in the user interface?
        """
        if value.defined_when == '':
            return True
        return ui.eval_when(value.defined_when)

    #-------------------------------------------------------------------------
    #  Parses Group options specified as a string:
    #-------------------------------------------------------------------------

    def _parse(self, value):
        """ Parses Group options specified as a string.
        """
        # Override the defaults, since we only allow 'True' values to be
        # specified:
        self.show_border = self.show_labels = self.show_left = False

        # Parse all of the single or multi-character options:
        value, empty = self._parse_label(value)
        value = self._parse_style(value)
        value = self._option(value, '-', 'orientation', 'horizontal')
        value = self._option(value, '|', 'orientation', 'vertical')
        value = self._option(value, '=', 'layout', 'split')
        value = self._option(value, '^', 'layout', 'tabbed')
        value = self._option(value, '>', 'show_labels', True)
        value = self._option(value, '<', 'show_left', True)
        value = self._option(value, '!', 'selected', True)

        show_labels = not (self.show_labels and self.show_left)
        self.show_left = not self.show_labels
        self.show_labels = show_labels

        # Parse all of the punctuation based sub-string options:
        value = self._split('id', value, ':', _py2to3.str_find, 0, 1)
        if value != '':
            self.object = value

    #-------------------------------------------------------------------------
    #  Handles a label being found in the string definition:
    #-------------------------------------------------------------------------

    def _parsed_label(self):
        """ Handles a label being found in the string definition.
        """
        self.show_border = True

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of the Group:
    #-------------------------------------------------------------------------

    def __repr__(self):
        """ Returns a "pretty print" version of the Group.
        """
        result = []
        items = ',\n'.join([item.__repr__() for item in self.content])
        if len(items) > 0:
            result.append(items)

        options = self._repr_options(
            'orientation',
            'show_border',
            'show_labels',
            'show_left',
            'selected',
            'id',
            'object',
            'label',
            'style',
            'layout',
            'style_sheet')
        if options is not None:
            result.append(options)

        content = ',\n'.join(result)
        if len(content) == 0:
            return self.__class__.__name__ + '()'

        return '%s(\n%s\n)' % (
               self.__class__.__name__, self._indent(content))

    #-------------------------------------------------------------------------
    #  Property getters/setters for width/height attributes
    #-------------------------------------------------------------------------

    @cached_property
    def _get_width(self):
        """ Returns the requested width of the Group.
        """
        width = 0.0
        for item in self.content:
            if item.width >= 1:
                if self.orientation == 'horizontal':
                    width += item.width
                elif self.orientation == 'vertical':
                    width = max(width, item.width)

        if width == 0:
            width = -1.0

        return width

    @cached_property
    def _get_height(self):
        """ Returns the requested height of the Group.
        """
        height = 0.0
        for item in self.content:
            if item.height >= 1:
                if self.orientation == 'horizontal':
                    height = max(height, item.height)
                elif self.orientation == 'vertical':
                    height += item.height

        if height == 0:
            height = -1.0

        return height

#-------------------------------------------------------------------------
#  'HGroup' class:
#-------------------------------------------------------------------------


class HGroup(Group):
    """ A group whose items are laid out horizontally.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it horizontal group
    # behavior:
    orientation = 'horizontal'

#-------------------------------------------------------------------------
#  'VGroup' class:
#-------------------------------------------------------------------------


class VGroup(Group):
    """ A group whose items are laid out vertically.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it vertical group
    # behavior:
    orientation = 'vertical'

#-------------------------------------------------------------------------
#  'VGrid' class:
#-------------------------------------------------------------------------


class VGrid(VGroup):
    """ A group whose items are laid out in 2 columns.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it grid behavior:
    columns = 2

#-------------------------------------------------------------------------
#  'HFlow' class:
#-------------------------------------------------------------------------


class HFlow(HGroup):
    """ A group in which items are laid out horizontally, and "wrap" when
    they exceed the available horizontal space..
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it horizontal flow
    # behavior:
    layout = 'flow'
    show_labels = False

#-------------------------------------------------------------------------
#  'VFlow' class:
#-------------------------------------------------------------------------


class VFlow(VGroup):
    """ A group in which items are laid out vertically, and "wrap" when they
    exceed the available vertical space.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it vertical flow behavior:
    layout = 'flow'
    show_labels = False

#-------------------------------------------------------------------------
#  'VFold' class:
#-------------------------------------------------------------------------


class VFold(VGroup):
    """ A group in which items are laid out vertically and can be collapsed
        (i.e. 'folded') by clicking their title.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it vertical folding group
    # behavior:
    layout = 'fold'
    show_labels = False

#-------------------------------------------------------------------------
#  'HSplit' class:
#-------------------------------------------------------------------------


class HSplit(Group):
    """ A horizontal group with splitter bars to separate it from other groups.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it horizontal splitter
    # behavior:
    layout = 'split'
    orientation = 'horizontal'

#-------------------------------------------------------------------------
#  'VSplit' class:
#-------------------------------------------------------------------------


class VSplit(Group):
    """ A vertical group with splitter bars to separate it from other groups.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it vertical splitter
    # behavior:
    layout = 'split'
    orientation = 'vertical'

#-------------------------------------------------------------------------
#  'Tabbed' class:
#-------------------------------------------------------------------------


class Tabbed(Group):
    """ A group that is shown as a tabbed notebook.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Override standard Group trait defaults to give it tabbed notebook
    # behavior:
    layout = 'tabbed'
    springy = True

#-------------------------------------------------------------------------
#  'ShadowGroup' class:
#-------------------------------------------------------------------------


class ShadowGroup(Group):
    """ Corresponds to a Group object, but with all embedded Include
        objects resolved, and with all embedded Group objects replaced by
        corresponding ShadowGroup objects.
    """

    def __init__(self, shadow, **traits):
        # Set the 'shadow' trait before all others, to avoid exceptions
        # when setting those other traits.
        self.shadow = shadow
        super(ShadowGroup, self).__init__(**traits)

    #-------------------------------------------------------------------------
    # Trait definitions:
    #-------------------------------------------------------------------------

    # Group object this is a "shadow" for
    shadow = ReadOnly

    # Number of ShadowGroups in **content**
    groups = ReadOnly

    # Name of the group
    id = ShadowDelegate

    # User interface label for the group
    label = ShadowDelegate

    # Default context object for group items
    object = ShadowDelegate

    # Default style of items in the group
    style = ShadowDelegate

    # Default docking style of items in the group
    dock = ShadowDelegate

    # Default image to display on notebook tabs
    image = ShadowDelegate

    # The theme to use for a DockWindow:
    dock_theme = ShadowDelegate

    # The theme to use for the group itself:
    group_theme = ShadowDelegate

    # The theme to use for item's contained in the group:
    item_theme = ShadowDelegate

    # The theme to use for the labels of items contained in the group:
    label_theme = ShadowDelegate

    # Category of elements dragged from the view
    export = ShadowDelegate

    # Spatial orientation of the group
    orientation = ShadowDelegate

    # Layout style of the group
    layout = ShadowDelegate

    # Should the group be scrollable along the direction of orientation?
    scrollable = ShadowDelegate

    # The number of columns in the group
    columns = ShadowDelegate

    # Should a border be drawn around group?
    show_border = ShadowDelegate

    # Should labels be added to items in group?
    show_labels = ShadowDelegate

    # Should labels be shown to the left of items (vs. the right)?
    show_left = ShadowDelegate

    # Is group the initially selected page?
    selected = ShadowDelegate

    # Should the group use extra space along its parent group's layout
    # orientation?
    springy = ShadowDelegate

    # Optional help text (for top-level group)
    help = ShadowDelegate

    # Pre-condition for defining the group
    defined_when = ShadowDelegate

    # Pre-condition for showing the group
    visible_when = ShadowDelegate

    # Pre-condition for enabling the group
    enabled_when = ShadowDelegate

    # Amount of padding to add around each item
    padding = ShadowDelegate

    # Style sheet for the panel
    style_sheet = ShadowDelegate

    #-------------------------------------------------------------------------
    #  Returns the contents of the ShadowGroup within a specified user interface
    #  building context. This makes sure that all Group types are of the same
    #  type (i.e. Group or Item) and that all Include objects have been replaced
    #  by their substituted values:
    #-------------------------------------------------------------------------

    def get_content(self, allow_groups=True):
        """ Returns the contents of the Group within a specified context for
        building a user interface.

        This method makes sure that all Group types are of the same type (i.e.,
        Group or Item) and that all Include objects have been replaced by their
        substituted values.
        """
        # Make a copy of the content:
        result = self.content[:]

        # If result includes any ShadowGroups and they are not allowed,
        # replace them:
        if self.groups != 0:
            if not allow_groups:
                i = 0
                while i < len(result):
                    value = result[i]
                    if isinstance(value, ShadowGroup):
                        items = value.get_content(False)
                        result[i:i + 1] = items
                        i += len(items)
                    else:
                        i += 1
            elif (self.groups != len(result)) and (self.layout == 'normal'):
                items = []
                content = []
                for item in result:
                    if isinstance(item, ShadowGroup):
                        self._flush_items(content, items)
                        content.append(item)
                    else:
                        items.append(item)
                self._flush_items(content, items)
                result = content

        # Return the resulting list of objects:
        return result

    #-------------------------------------------------------------------------
    #  Returns an id used to identify the group:
    #-------------------------------------------------------------------------

    def get_id(self):
        """ Returns an ID for the group.
        """
        if self.id != '':
            return self.id

        return ':'.join([item.get_id() for item in self.get_content()])

    #-------------------------------------------------------------------------
    #  Sets the correct container for the content:
    #-------------------------------------------------------------------------

    def set_container(self):
        """ Sets the correct container for the content.
        """
        pass

    #-------------------------------------------------------------------------
    #  Creates a sub-Group for any items contained in a specified list:
    #-------------------------------------------------------------------------

    def _flush_items(self, content, items):
        """ Creates a sub-group for any items contained in a specified list.
        """
        if len(items) > 0:
            content.append(
                # Set shadow before hand to prevent delegation errors
                ShadowGroup(shadow=self.shadow).set(
                    groups=0,
                    label='',
                    show_border=False,
                    content=items,
                    show_labels=self.show_labels,
                    show_left=self.show_left,
                    springy=self.springy,
                    orientation=self.orientation
                ))
            del items[:]

    #-------------------------------------------------------------------------
    #  Returns a 'pretty print' version of the Group:
    #-------------------------------------------------------------------------

    def __repr__(self):
        """ Returns a "pretty print" version of the Group.
        """
        return repr(self.shadow)



# Standard library imports.
import os

# Enthought library imports.
from traits.api import HasTraits, on_trait_change, Any, Instance, List, Bool, Int, Range, Str, Unicode, Event
from apptools.preferences.api import PreferencesHelper
from envisage.ui.tasks.api import PreferencesPane, TaskExtension
from envisage.api import ExtensionPoint, Plugin
from pyface.action.api import Action, ActionItem, Group, Separator
from pyface.tasks.action.api import SMenu, TaskAction, EditorAction, SchemaAddition
from traitsui.api import EnumEditor, HGroup, VGroup, Item, Label, View

from omnivore.framework.plugin import FrameworkPlugin
from omnivore.framework.actions import ApplicationDynamicSubmenuGroup

import logging
log = logging.getLogger(__name__)


class OpenRecentPreferences(PreferencesHelper):
    """ The preferences helper for the Framework application.
    """

    #### 'PreferencesHelper' interface ########################################

    # The path to the preference node that contains the preferences.
    preferences_path = 'open_recent'

    #### Preferences ##########################################################

    # Controls where the new files are added: top or bottom
    add_at_top = Bool(True)

    list_max = 50 # not a trait, can't seem to programmically pull out the max value of the range object

    # Maximum number of files to retain in the list
    list_length = Range(low=4, high=list_max, value=10)

    def max_list_length(self):
        # No easy way to get the maximum value of a Range object
        t = self.trait('list_length')
        values = t.full_info(None, None, None).split()
        return int(values[-1])


class OpenRecentPreferencesPane(PreferencesPane):
    """ The preferences pane for the Framework application.
    """

    #### 'PreferencesPane' interface ##########################################

    # The factory to use for creating the preferences model object.
    model_factory = OpenRecentPreferences

    #### 'FrameworkPreferencesPane' interface ################################

    view = View(
        VGroup(HGroup(Item('list_length'),
                      Label('Number of recent files to remember'),
                      show_labels = False),
               HGroup(Item('add_at_top'),
                      Label('Add new file to top of list'),
                      show_labels = False),
               label='Open Recent'),
        resizable=True)

# FIXME: RecentFiles doesn't work as a traits object! Weird.
#
#class RecentFiles(HasTraits):
#    """Open a file from the list of recently opened files.
#
#    Maintains a list of the most recently opened files so that the next time
#    you start the application you can see what you were editing last time.
#    This list is automatically maintained, so every time you open a new file,
#    it is added to the list.  This list is limited in size by the classpref
#    'list_length' in L{RecentFilesPlugin}.
#    """
#
#    # Preferences aren't duplicated here; instead, reference the helper so they
#    # can be lookup up every time they are needed.
#    helper = Instance(OpenRecentPreferences)
#
#    # location to serialize the list to persist across application runs
#    serialize_uri = Unicode
#
#    storage_flag = Bool
#
#    updated = Event


class RecentFiles(object):
    """Open a file from the list of recently opened files.
    
    Maintains a list of the most recently opened files so that the next time
    you start the application you can see what you were editing last time.
    This list is automatically maintained, so every time you open a new file,
    it is added to the list.  This list is limited in size by the classpref
    'list_length' in L{RecentFilesPlugin}.
    """

    def __init__(self, helper, serialize_uri):
        self.helper = helper
        self.serialize_uri = serialize_uri
        self._storage_flag_default()

    #### Trait initializers ###################################################

    def _storage_flag_default(self):
        self.storage_flag = True
        self.storage = []
        log.debug("STORAGE_FLAG!!! %s" % self.storage_flag)
        self.unserialize()

    def is_acceptable_uri(self, uri):
        # skip files with the about: protocol
        #return uri.scheme != 'about' and uri.scheme != 'mem'
        return True

    def iter_items(self):
        for item in self.storage[0:self.helper.list_length]:
            yield item

    def serialize(self):
        """Serialize the current items to the file"""
        log.debug("SERIALIZING: %s" % str(self.storage))
        # trim list to maximum possible length
        self.storage[self.helper.max_list_length():] = []
        with open(self.serialize_uri,'w') as fh:
            try:
                for item in self.iter_items():
                    fh.write("%s%s" % (item.encode('utf8'),os.linesep))
            except:
                pass

    def unserialize(self):
        """Unserialize items from the file into a list"""
        log.debug("UNSERIALIZING: %s" % str(self.serialize_uri))
        self.storage = []
        try:
            with open(self.serialize_uri,'r') as fh:
                for line in fh:
                    trimmed = line.decode('utf8').rstrip()
                    if trimmed.strip():
                        self.storage.append(trimmed)
        except:
            pass

    def append_uri(self, uri, extra=None):
        if self.is_acceptable_uri(uri):
            item = unicode(uri)
            if extra:
                item = (item, extra)
            # if we're adding an item that's already in the list, move it
            # to the top of the list by recreating the list
            try:
                index = self.storage.index(item)
                self.storage.pop(index)
            except ValueError:
                pass
            # New items are added at the top of this list
            self.storage[0:0]=[item]
            self.serialize()
        log.debug("STORAGE: %s" % str(self.storage))


class OpenRecentAction(Action):
    """This submenu contain a list of the files most recently loaded or saved.

    You can limit the number of items to remember in the General tab of the
    `Preferences`_ dialog.
    """
    doc_hint = "summary"

    uri = Unicode

    def _name_default(self):
        return self.uri

    def perform(self, event):
        event.task.window.application.load_file(self.uri, event.task)


class OpenRecentGroup(ApplicationDynamicSubmenuGroup):
    """ A menu for creating a new file for each type of task
    """

    #### 'ActionManager' interface ############################################

    id = 'NewFileGroup'

    event_name = 'plugin_event'

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        items = []
        recent_files = self.application.get_plugin_data('open_recent')
        log.debug("GET ITEMS: recent_files=%s, storage=%s" % (recent_files, recent_files.storage))
        if recent_files is not None:
            for uri in recent_files.iter_items():
                action = OpenRecentAction(uri=uri)
                items.append(ActionItem(action=action))

        return items


class OpenRecentPlugin(FrameworkPlugin):

#    recent_files = None

    #### 'IPlugin' interface ##################################################

    # The plugin's unique identifier.
    id = 'open_recent'

    # The plugin's name (suitable for displaying to the user).
    name = 'Recently Opened Files List'

    # Extension point IDs.
    PREFERENCES_PANES = 'envisage.ui.tasks.preferences_panes'
    TASK_EXTENSIONS   = 'envisage.ui.tasks.task_extensions'
    OSX_MINIMAL_MENU = 'omnivore.osx_minimal_menu'

    #### Contributions to extension points made by this plugin ################

    preferences_panes = List(contributes_to=PREFERENCES_PANES)
    actions = List(contributes_to=TASK_EXTENSIONS)
    osx_actions = List(contributes_to=OSX_MINIMAL_MENU)

    def _osx_actions_default(self):
        submenu = lambda: SMenu(
            id='OpenRecentSubmenu', name="Open Recent"
        )
        actions = [
            SchemaAddition(factory=submenu,
                           path='MenuBar/File',
                           after="OpenGroup", before="OpenGroupEnd",
                           ),
            SchemaAddition(id='OpenRecent',
                           factory=OpenRecentGroup,
                           path='MenuBar/File/OpenRecentSubmenu'),
            ]

        return actions

    def _actions_default(self):
        # Using the same actions as the minimal menu
        actions = self._osx_actions_default()
        return [ TaskExtension(actions=actions) ]

    def _preferences_panes_default(self):
        return [ OpenRecentPreferencesPane ]

    def start(self):
        helper = self.get_helper(OpenRecentPreferences)

        # self.home is the config directory created especially for this plugin!
        recent_files = RecentFiles(helper=helper,
                                   serialize_uri=os.path.join(self.home, "files.dat"))
        self.set_plugin_data(recent_files)

    @on_trait_change('application:successfully_loaded_event')
    def update_recent_file(self, uri):
        log.debug("NEWLY LOADED FILE: %s" % uri)
        recent_files = self.get_plugin_data()
        try:
            recent_files.append_uri(uri)
        except Exception, e:
            log.warning("FAILED ADDING %s to recent files list: %s" % (uri, e.message))
            return
        self.fire_plugin_event()



# -*- encoding: utf-8 -*-
"""

  fs.tests.test_errors:  testcases for the fs error classes functions

"""


import unittest
import fs.tests
from fs.errors import *
import pickle

from fs.path import *

class TestErrorPickling(unittest.TestCase):

    def test_pickling(self):
        def assert_dump_load(e):
            e2 = pickle.loads(pickle.dumps(e))
            self.assertEqual(e.__dict__,e2.__dict__)
        assert_dump_load(FSError())
        assert_dump_load(PathError("/some/path"))
        assert_dump_load(ResourceNotFoundError("/some/other/path"))
        assert_dump_load(UnsupportedError("makepony"))


class TestFSError(unittest.TestCase):

    def test_unicode_representation_of_error_with_non_ascii_characters(self):
        path_error = PathError('/Shïrê/Frødø')
        _ = unicode(path_error)


#------------------------------------------------------------------------------
# Copyright (c) 2016, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" The interface for a dialog that prompts for a choice from a list. """


# Enthought library imports.
from traits.api import Any, List, Str

# Local imports.
from pyface.i_dialog import IDialog


class ISingleChoiceDialog(IDialog):
    """ The interface for a dialog that prompts for a choice from a list. """

    #### 'ISingleChoiceDialog' interface ######################################

    #: List of objects to choose from.
    choices = List(Any)

    #: The object chosen, if any.
    choice = Any

    #: An optional attribute to use for the name of each object in the dialog.
    name_attribute = Str

    #: The message to display to the user.
    message = Str


class MSingleChoiceDialog(object):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IConfirmationDialog interface.
    """

    def _choice_strings(self):
        """ Returns the list of strings to display in the dialog. """
        choices = self.choices
        if self.name_attribute != '':
            # choices is a list of objects with this attribute
            choices = [getattr(obj, self.name_attribute) for obj in choices]

        choices = [str(obj) for obj in choices]

        if len(choices) == 0:
            raise ValueError("SingleChoiceDialog requires at least 1 choice.")
        elif len(choices) != len(set(choices)):
            raise ValueError("Dialog choices {} contain repeated string value."
                             % choices)
        return choices



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Standard library imports.
import cPickle as pickle
import errno
import os

# Enthought library imports.
from traits.etsconfig.api import ETSConfig


class Persistent(object):
    """This persists a Traits class to a file.  It is used by the default
    permissions policy and user manager to implement basic (ie.  insecure)
    shared data storage."""

    def __init__(self, factory, file_name, desc):
        """Initialise the object.  factory is a callable that will create a
        new instance if there is no existing data.  file_name is the name of
        the file in the ETSConfig.application_home directory (or the directory
        specified by the ETS_PERMS_DATA_DIR environment variable if set) to
        persist the data to.  desc is a description of the data used in
        exceptions."""

        self._factory = factory
        self._desc = desc

        # Get the name of the file to use.
        data_dir = os.environ.get('ETS_PERMS_DATA_DIR',
                ETSConfig.application_home)
        self._fname = os.path.join(data_dir, file_name)
        self._lock = self._fname + '.lock'

        try:
            os.makedirs(data_dir)
        except OSError:
            pass

    def lock(self):
        """Obtain a lock on the persisted data."""

        try:
            os.mkdir(self._lock)
        except OSError, e:
            if e.errno == errno.EEXIST:
                msg = "The lock on %s is held by another application or user." % self._desc
            else:
                msg = "Unable to acquire lock on %s: %s." % (self._desc, e)

            raise PersistentError(msg)

    def unlock(self):
        """Release the lock on the persisted data."""

        try:
            os.rmdir(self._lock)
        except OSError, e:
            raise PersistentError("Unable to release lock on %s: %s." % (self._desc, e))

    def read(self):
        """Read and return the persisted data."""

        try:
            f = open(self._fname, 'r')

            try:
                try:
                    data = pickle.load(f)
                except:
                    raise PersistentError("Unable to read %s." % self._desc)
            finally:
                f.close()
        except IOError, e:
            if e.errno == errno.ENOENT:
                data = self._factory()
            else:
                raise PersistentError("Unable to open %s: %s." % (self._desc, e))

        return data

    def write(self, data):
        """Write the persisted data."""

        try:
            f = open(self._fname, 'w')

            try:
                try:
                    pickle.dump(data, f)
                except:
                    raise PersistentError("Unable to write %s." % self._desc)
            finally:
                f.close()
        except IOError, e:
            raise PersistentError("Unable to create %s: %s." % (self._desc, e))


class PersistentError(Exception):
    """The class used for all persistence related exceptions."""



""" Tests for extension point bindings. """


# Enthought library imports.
from envisage.api import ExtensionPoint
from envisage.api import bind_extension_point
from traits.api import HasTraits, List
from traits.testing.unittest_tools import unittest

# Local imports.
#
# We do these as absolute imports to allow nose to run from a different
# working directory.
from envisage.tests.mutable_extension_registry import (
    MutableExtensionRegistry
)


def listener(obj, trait_name, old, new):
    """ A useful trait change handler for testing! """

    listener.obj = obj
    listener.trait_name = trait_name
    listener.old = old
    listener.new = new

    return


class ExtensionPointBindingTestCase(unittest.TestCase):
    """ Tests for extension point binding. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.extension_registry = MutableExtensionRegistry()

        # Use the extension registry for all extension points and bindings.
        ExtensionPoint.extension_registry = self.extension_registry

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_untyped_extension_point(self):
        """ untyped extension point """

        registry = self.extension_registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Add an extension.
        registry.add_extension('my.ep', 42)

        # Declare a class that consumes the extension.
        class Foo(HasTraits):
            x = List

        f = Foo()
        f.on_trait_change(listener)

        # Make some bindings.
        bind_extension_point(f, 'x', 'my.ep')

        # Make sure that the object was initialized properly.
        self.assertEqual(1, len(f.x))
        self.assertEqual(42, f.x[0])

        # Add another extension.
        registry.add_extension('my.ep', 'a string')

        # Make sure that the object picked up the new extension...
        self.assertEqual(2, len(f.x))
        self.assert_(42 in f.x)
        self.assert_('a string' in f.x)

        # ... and that the correct trait change event was fired.
        self.assertEqual(f, listener.obj)
        self.assertEqual('x_items', listener.trait_name)
        self.assertEqual(1, len(listener.new.added))
        self.assert_('a string' in listener.new.added)

        return

    def test_set_extensions_via_trait(self):
        """ set extensions via trait """

        registry = self.extension_registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Add an extension.
        registry.add_extension('my.ep', 42)

        # Declare a class that consumes the extension.
        class Foo(HasTraits):
            x = List

        f = Foo()
        f.on_trait_change(listener)

        # Make some bindings.
        bind_extension_point(f, 'x', 'my.ep')

        # Make sure that the object was initialized properly.
        self.assertEqual(1, len(f.x))
        self.assertEqual(42, f.x[0])

        # Set the extensions.
        f.x = ['a string']

        # Make sure that the object picked up the new extension...
        self.assertEqual(1, len(f.x))
        self.assert_('a string' in f.x)

        self.assertEqual(1, len(registry.get_extensions('my.ep')))
        self.assert_('a string' in registry.get_extensions('my.ep'))

        # ... and that the correct trait change event was fired.
        self.assertEqual(f, listener.obj)
        self.assertEqual('x', listener.trait_name)
        self.assertEqual(1, len(listener.new))
        self.assert_('a string' in listener.new)

        return

    def test_set_extensions_via_registry(self):
        """ set extensions via registry """

        registry = self.extension_registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Add an extension.
        registry.add_extension('my.ep', 42)

        # Declare a class that consumes the extension.
        class Foo(HasTraits):
            x = List

        f = Foo()
        f.on_trait_change(listener)

        # Make some bindings.
        bind_extension_point(f, 'x', 'my.ep')

        # Make sure that the object was initialized properly.
        self.assertEqual(1, len(f.x))
        self.assertEqual(42, f.x[0])

        # Set the extensions.
        registry.set_extensions('my.ep', ['a string'])

        # Make sure that the object picked up the new extension...
        self.assertEqual(1, len(f.x))
        self.assert_('a string' in f.x)

        # ... and that the correct trait change event was fired.
        self.assertEqual(f, listener.obj)
        self.assertEqual('x', listener.trait_name)
        self.assertEqual(1, len(listener.new))
        self.assert_('a string' in listener.new)

        return

    def test_explicit_extension_registry(self):
        """ explicit extension registry """

        registry = self.extension_registry

        # Add an extension point.
        registry.add_extension_point(self._create_extension_point('my.ep'))

        # Add an extension.
        registry.add_extension('my.ep', 42)

        # Declare a class that consumes the extension.
        class Foo(HasTraits):
            x = List

        f = Foo()
        f.on_trait_change(listener)

        # Create an empty extension registry use that in the binding.
        extension_registry = MutableExtensionRegistry()

        # Make some bindings.
        bind_extension_point(f, 'x', 'my.ep', extension_registry)

        # Make sure that we pick up the empty extension registry and not the
        # default one.
        self.assertEqual(0, len(f.x))

        return

    def test_should_be_able_to_bind_multiple_traits_on_a_single_object(self):

        registry = self.extension_registry

        # Add 2 extension points.
        registry.add_extension_point(self._create_extension_point('my.ep'))
        registry.add_extension_point(self._create_extension_point('another.ep'))

        # Declare a class that consumes both of the extension points.
        class Foo(HasTraits):
            x = List
            y = List

        f = Foo()

        # Bind two different traits on the object to the extension points.
        bind_extension_point(f, 'x', 'my.ep', registry)
        bind_extension_point(f, 'y', 'another.ep', registry)
        self.assertEqual(0, len(f.x))
        self.assertEqual(0, len(f.y))

        # Add some contributions to the extension points.
        registry.add_extension('my.ep', 42)
        registry.add_extensions('another.ep', [98, 99, 100])

        # Make sure both traits were bound correctly.
        self.assertEqual(1, len(f.x))
        self.assertEqual(3, len(f.y))

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_extension_point(self, id, trait_type=List, desc=''):
        """ Create an extension point. """

        return ExtensionPoint(id=id, trait_type=trait_type, desc=desc)


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



from .application_browser_view import ApplicationBrowserView
from .extension_registry_browser_view import ExtensionRegistryBrowserView
from .service_registry_browser_view import ServiceRegistryBrowserView



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
from __future__ import absolute_import

from .column_provider import ColumnProvider
from .content_provider import ContentProvider
from .content_viewer import ContentViewer
from .default_tree_content_provider import DefaultTreeContentProvider
from .label_provider import LabelProvider
from .table_column_provider import TableColumnProvider
from .table_content_provider import TableContentProvider
from .table_label_provider import TableLabelProvider
from .tree_content_provider import TreeContentProvider
from .tree_label_provider import TreeLabelProvider
from .tree_item import TreeItem
from .viewer import Viewer
from .viewer_filter import ViewerFilter
from .viewer_sorter import ViewerSorter

# these are only implemented in wx at the moment
from .table_viewer import TableViewer
from .tree_viewer import TreeViewer



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A reference to an object that lives outside of the naming system. """


# Enthought library imports.
from traits.api import Any, HasPrivateTraits, List, Str

# Local imports.
from address import Address


class Reference(HasPrivateTraits):
    """ A reference to an object that lives outside of the naming system.

    References provide a way to store the address(s) of objects that live
    outside of the naming system.  A reference consists of a list of
    addresses that represent a communications endpoint for the object being
    referenced.

    A reference also contains information to assist in the creation of an
    instance of the object to which it refers.  It contains the name of
    the class that will be created and the class name and location of a
    factory that will be used to do the actual instance creation.

    """

    #### 'Reference' interface ################################################

    # The list of addresses that can be used to 'contact' the object.
    addresses = List(Address)

    # The class name of the object that this reference refers to.
    class_name = Str

    # The class name of the object factory.
    factory_class_name = Str

#### EOF ######################################################################



from traits.api import HasTraits, provides

from omnivore.file_type.i_file_recognizer import IFileRecognizer, RecognizerBase
from omnivore.utils.textutil import guessBinary


@provides(IFileRecognizer)
class PlainTextRecognizer(RecognizerBase):
    """Default plain text identifier based on percentage of non-ASCII bytes.
    
    """
    id = "text/plain"

    def identify(self, guess):
        if not guessBinary(guess.get_utf8()):
            return "text/plain"


@provides(IFileRecognizer)
class XMLTextRecognizer(RecognizerBase):
    """Default plain text identifier based on percentage of non-ASCII bytes.
    
    """
    id = "text/xml"

    before = "text/plain"

    def identify(self, guess):
        byte_stream = guess.get_utf8().strip()
        if not byte_stream.startswith("<"):
            return
        if byte_stream.startswith("<?xml"):
            found_xml = True
        else:
            found_xml = False
        if "<rss" in byte_stream:
            return "application/rss+xml"
        byte_stream = byte_stream.lower()
        if "<!doctype html" in byte_stream or "<html" in byte_stream:
            return "text/html"

        if found_xml:
            return "text/xml"


@provides(IFileRecognizer)
class PoundBangTextRecognizer(RecognizerBase):
    """Recognizer for pound bang style executable script files
    
    """
    id = "text/poundbang"

    before = "text/plain"

    def identify(self, guess):
        byte_stream = guess.get_utf8()
        if not byte_stream.startswith("#!"):
            return
        line = byte_stream[2:80].lower().strip()
        if line.startswith("/usr/bin/env"):
            line = line[12:].strip()
        words = line.split()
        names = words[0].split("/")
        if names[-1]:
            return "text/%s" % names[-1]



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# The script manager.
_script_manager = None


def get_script_manager():
    """Return the IScriptManager implementation, creating a ScriptManager
    instance if no other implementation has been set."""

    global _script_manager

    if _script_manager is None:
        from script_manager import ScriptManager

        _script_manager = ScriptManager()

    return _script_manager


def set_script_manager(script_manager):
    """Set the IScriptManager implementation to use."""

    global _script_manager

    _script_manager = script_manager



"""
fs.expose.xmlrpc
================

Server to expose an FS via XML-RPC

This module provides the necessary infrastructure to expose an FS object
over XML-RPC.  The main class is 'RPCFSServer', a SimpleXMLRPCServer subclass
designed to expose an underlying FS.

If you need to use a more powerful server than SimpleXMLRPCServer, you can
use the RPCFSInterface class to provide an XML-RPC-compatible wrapper around
an FS object, which can then be exposed using whatever server you choose
(e.g. Twisted's XML-RPC server).

"""

import xmlrpclib
from SimpleXMLRPCServer import SimpleXMLRPCServer
from datetime import datetime
import base64

import six
from six import PY3


class RPCFSInterface(object):
    """Wrapper to expose an FS via a XML-RPC compatible interface.

    The only real trick is using xmlrpclib.Binary objects to transport
    the contents of files.
    """

    # info keys are restricted to a subset known to work over xmlrpc
    # This fixes an issue with transporting Longs on Py3
    _allowed_info = ["size",
                     "created_time",
                     "modified_time",
                     "accessed_time",
                     "st_size",
                     "st_mode",
                     "type"]

    def __init__(self, fs):
        super(RPCFSInterface, self).__init__()
        self.fs = fs

    def encode_path(self, path):
        """Encode a filesystem path for sending over the wire.

        Unfortunately XMLRPC only supports ASCII strings, so this method
        must return something that can be represented in ASCII.  The default
        is base64-encoded UTF-8.
        """
        #return path
        return six.text_type(base64.b64encode(path.encode("utf8")), 'ascii')

    def decode_path(self, path):
        """Decode paths arriving over the wire."""
        return six.text_type(base64.b64decode(path.encode('ascii')), 'utf8')

    def getmeta(self, meta_name):
        meta = self.fs.getmeta(meta_name)
        if isinstance(meta, basestring):
            meta = self.decode_path(meta)
        return meta

    def getmeta_default(self, meta_name, default):
        meta = self.fs.getmeta(meta_name, default)
        if isinstance(meta, basestring):
            meta = self.decode_path(meta)
        return meta

    def hasmeta(self, meta_name):
        return self.fs.hasmeta(meta_name)

    def get_contents(self, path, mode="rb"):
        path = self.decode_path(path)
        data = self.fs.getcontents(path, mode)
        return xmlrpclib.Binary(data)

    def set_contents(self, path, data):
        path = self.decode_path(path)
        self.fs.setcontents(path, data.data)

    def exists(self, path):
        path = self.decode_path(path)
        return self.fs.exists(path)

    def isdir(self, path):
        path = self.decode_path(path)
        return self.fs.isdir(path)

    def isfile(self, path):
        path = self.decode_path(path)
        return self.fs.isfile(path)

    def listdir(self, path="./", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        path = self.decode_path(path)
        entries = self.fs.listdir(path, wildcard, full, absolute, dirs_only, files_only)
        return [self.encode_path(e) for e in entries]

    def makedir(self, path, recursive=False, allow_recreate=False):
        path = self.decode_path(path)
        return self.fs.makedir(path, recursive, allow_recreate)

    def remove(self, path):
        path = self.decode_path(path)
        return self.fs.remove(path)

    def removedir(self, path, recursive=False, force=False):
        path = self.decode_path(path)
        return self.fs.removedir(path, recursive, force)

    def rename(self, src, dst):
        src = self.decode_path(src)
        dst = self.decode_path(dst)
        return self.fs.rename(src, dst)

    def settimes(self, path, accessed_time, modified_time):
        path = self.decode_path(path)
        if isinstance(accessed_time, xmlrpclib.DateTime):
            accessed_time = datetime.strptime(accessed_time.value, "%Y%m%dT%H:%M:%S")
        if isinstance(modified_time, xmlrpclib.DateTime):
            modified_time = datetime.strptime(modified_time.value, "%Y%m%dT%H:%M:%S")
        return self.fs.settimes(path, accessed_time, modified_time)

    def getinfo(self, path):
        path = self.decode_path(path)
        info = self.fs.getinfo(path)
        info = dict((k, v) for k, v in info.iteritems()
                    if k in self._allowed_info)
        return info

    def desc(self, path):
        path = self.decode_path(path)
        return self.fs.desc(path)

    def getxattr(self, path, attr, default=None):
        path = self.decode_path(path)
        attr = self.decode_path(attr)
        return self.fs.getxattr(path, attr, default)

    def setxattr(self, path, attr, value):
        path = self.decode_path(path)
        attr = self.decode_path(attr)
        return self.fs.setxattr(path, attr, value)

    def delxattr(self, path, attr):
        path = self.decode_path(path)
        attr = self.decode_path(attr)
        return self.fs.delxattr(path, attr)

    def listxattrs(self, path):
        path = self.decode_path(path)
        return [self.encode_path(a) for a in self.fs.listxattrs(path)]

    def copy(self, src, dst, overwrite=False, chunk_size=16384):
        src = self.decode_path(src)
        dst = self.decode_path(dst)
        return self.fs.copy(src, dst, overwrite, chunk_size)

    def move(self, src, dst, overwrite=False, chunk_size=16384):
        src = self.decode_path(src)
        dst = self.decode_path(dst)
        return self.fs.move(src, dst, overwrite, chunk_size)

    def movedir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        src = self.decode_path(src)
        dst = self.decode_path(dst)
        return self.fs.movedir(src, dst, overwrite, ignore_errors, chunk_size)

    def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        src = self.decode_path(src)
        dst = self.decode_path(dst)
        return self.fs.copydir(src, dst, overwrite, ignore_errors, chunk_size)


class RPCFSServer(SimpleXMLRPCServer):
    """Server to expose an FS object via XML-RPC.

    This class takes as its first argument an FS instance, and as its second
    argument a (hostname,port) tuple on which to listen for XML-RPC requests.
    Example::

        fs = OSFS('/var/srv/myfiles')
        s = RPCFSServer(fs,("",8080))
        s.serve_forever()

    To cleanly shut down the server after calling serve_forever, set the
    attribute "serve_more_requests" to False.
    """

    def __init__(self, fs, addr, requestHandler=None, logRequests=None):
        kwds = dict(allow_none=True)
        if requestHandler is not None:
            kwds['requestHandler'] = requestHandler
        if logRequests is not None:
            kwds['logRequests'] = logRequests
        self.serve_more_requests = True
        SimpleXMLRPCServer.__init__(self, addr, **kwds)
        self.register_instance(RPCFSInterface(fs))

    def serve_forever(self):
        """Override serve_forever to allow graceful shutdown."""
        while self.serve_more_requests:
            self.handle_request()



""" A Python based configuration file with hierarchical sections. """


class PyConfigFile(dict):
    """ A Python based configuration file with hierarchical sections. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, file_or_filename=None):
        """ Constructor.

        If 'file_or_filename' is specified it will be loaded immediately. It
        can be either:-

        a) a filename
        b) a file-like object that must be open for reading

        """

        # A dictionary containing one namespace instance for each root of the
        # config hierarchy (see the '_Namespace' class for more details).
        #
        # e.g. If the following sections have been loaded:-
        #
        # [acme.foo]
        # ...
        # [acme.bar]
        # ...
        # [tds]
        # ...
        # [tds.baz]
        # ...
        #
        # Then the dictionary will contain:-
        #
        # {'acme' : <A _Namespace instance>, 'tds' : <A _Namespace instance>}
        #
        self._namespaces = {}

        if file_or_filename is not None:
            self.load(file_or_filename)

        return

    ###########################################################################
    # 'PyConfigFile' interface.
    ###########################################################################

    def load(self, file_or_filename):
        """ Load the configuration from a file.

        'file_or_filename' can be either:-

        a) a filename
        b) a file-like object that must be open for reading

        """

        # Get an open file to read from.
        f = self._get_file(file_or_filename)

        section_name = None
        for line in f:
            stripped = line.strip()

            # Is this line a section header?
            #
            # If so then parse the preceding section (if there is one) and
            # start collecting the body of the new section.
            if stripped.startswith('[') and stripped.endswith(']'):
                if section_name is not None:
                    self._parse_section(section_name, section_body)

                section_name = stripped[1:-1]
                section_body = ''

            # Otherwise, this is *not* a section header so add the line to the
            # body of the current section. If there is no current section then
            # we simply ignore it!
            else:
                if section_name is not None:
                    section_body += line

        # Parse the last section in the file.
        if section_name is not None:
            self._parse_section(section_name, section_body)

        f.close()

        return

    def save(self, file_or_filename):
        """ Save the configuration to a file.

        'file_or_filename' can be either:-

        a) a filename
        b) a file-like object that must be open for writing

        """

        f = self._get_file(file_or_filename, 'w')

        for section_name, section_data in self.items():
            self._write_section(f, section_name, section_data)

        f.close()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_file(self, file_or_filename, mode='r'):
        """ Return an open file object from a file or a filename.

        The mode is only used if a filename is specified.

        """

        if isinstance(file_or_filename, basestring):
            f = open(file_or_filename, mode)

        else:
            f = file_or_filename

        return f

    def _get_namespace(self, section_name):
        """ Return the namespace that represents the section. """

        components = section_name.split('.')
        namespace = self._namespaces.setdefault(components[0], _Namespace())

        for component in components[1:]:
            namespace = getattr(namespace, component)

        return namespace

    def _parse_section(self, section_name, section_body):
        """ Parse a section.

        In this implementation, we don't actually 'parse' anything - we just
        execute the body of the section as Python code ;^)

        """

        # If this is the first time that we have come across the section then
        # start with an empty dictionary for its contents. Otherwise, we will
        # update its existing contents.
        section = self.setdefault(section_name, {})

        # Execute the Python code in the section dictionary.
        #
        # We use 'self._namespaces' as the globals for the code execution so
        # that config values can refer to other config values using familiar
        # Python syntax (see the '_Namespace' class for more details).
        #
        # e.g.
        #
        # [acme.foo]
        # bar = 1
        # baz = 99
        #
        # [acme.blargle]
        # blitzel = acme.foo.bar + acme.foo.baz
        exec section_body in self._namespaces, section

        # The '__builtins__' dictionary gets added to 'self._namespaces' as
        # by the call to 'exec'. However, we want 'self._namespaces' to only
        # contain '_Namespace' instances, so we do the cleanup here.
        del self._namespaces['__builtins__']

        # Get the section's corresponding node in the 'dotted' namespace and
        # update it with the config values.
        namespace = self._get_namespace(section_name)
        namespace.__dict__.update(section)

        return

    def _write_section(self, f, section_name, section_data):
        """ Write a section to a file. """

        f.write('[%s]\n' % section_name)

        for name, value in section_data.items():
            f.write('%s = %s\n' % (name, repr(value)))

        f.write('\n')

        return

    ###########################################################################
    # Debugging interface.
    ###########################################################################

    def _pretty_print_namespaces(self):
        """ Pretty print the 'dotted' namespaces. """

        for name, value in self._namespaces.items():
            print 'Namespace:', name
            value.pretty_print('  ')

        return


###############################################################################
# Internal use only.
###############################################################################

class _Namespace(object):
    """ An object that represents a node in a dotted namespace.

    We build up a dotted namespace so that config values can refer to other
    config values using familiar Python syntax.

    e.g.

    [acme.foo]
    bar = 1
    baz = 99

    [acme.blargle]
    blitzel = acme.foo.bar + acme.foo.baz

    """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __getattr__(self, name):
        """ Return the attribute with the specified name. """

        # This looks a little weird, but we are simply creating the next level
        # in the namespace hierarchy 'on-demand'.
        namespace = self.__dict__[name] = _Namespace()

        return namespace

    ###########################################################################
    # Debugging interface.
    ###########################################################################

    def pretty_print(self, indent=''):
        """ Pretty print the namespace. """

        for name, value in self.__dict__.items():
            if isinstance(value, _Namespace):
                print indent, 'Namespace:', name
                value.pretty_print(indent + '  ')

            else:
                print indent, name, ':', value

        return

#### EOF ######################################################################



#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt

from traitsui.ui_editors.data_frame_editor \
    import _DataFrameEditor as BaseDataFrameEditor

from .ui_editor import UIEditor


class _DataFrameEditor(BaseDataFrameEditor, UIEditor):
    """ Wx Toolkit implementation of the DataFrameEditor """
    pass



# Import the toolkit specific version.
from pyface.toolkit import toolkit_object
AdvancedEditorAreaPane = toolkit_object('tasks.advanced_editor_area_pane:'
                                        'AdvancedEditorAreaPane')



from traits.api import HasTraits, provides

from omnivore.file_type.i_file_recognizer import IFileRecognizer, RecognizerBase


@provides(IFileRecognizer)
class OmnivoreRecognizer(RecognizerBase):
    """Meta recognizer that forces the loading of a disk image when the
    .omnivore file is selected instead
    """

    name = "Omnivore Extra Metadata File"

    id = "text/vnd.omnivore.extra_metadata"

    before = "application/vnd.*"

    def identify(self, guess):
        uri = guess.metadata.uri
        if uri.endswith(".omnivore"):
            b = guess.bytes
            if b.startswith("#") or b.startswith("{"):
                uri = uri[0:-9]
                guess.reload(uri)



""" Context adapter for Python tuples. """


# Enthought library imports.
from traits.api import Tuple

# Local imports.
from list_context_adapter import ListContextAdapter


class TupleContextAdapter(ListContextAdapter):
    """ Context adapter for Python tuples. """

    #### 'ContextAdapter' interface ###########################################

    # The object that we are adapting.
    adaptee = Tuple

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2006, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/25/2006
#
#------------------------------------------------------------------------------

""" Defines the various editors for a drag-and-drop editor,
    for the wxPython user interface toolkit. A drag-and-drop editor represents
    its value as a simple image which, depending upon the editor style, can be
    a drag source only, a drop target only, or both a drag source and a drop
    target.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx
import numpy

from cPickle \
    import load

from traits.api \
    import Bool

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.dnd_editor file.
from traitsui.editors.dnd_editor \
    import ToolkitEditorFactory

from pyface.wx.drag_and_drop \
    import PythonDropSource, PythonDropTarget, clipboard

try:
    from apptools.io import File
except ImportError:
    File = None

try:
    from apptools.naming.api import Binding
except ImportError:
    Binding = None

from pyface.image_resource \
    import ImageResource

from editor \
    import Editor

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# The image to use when the editor accepts files:
file_image = ImageResource('file').create_image()

# The image to use when the editor accepts objects:
object_image = ImageResource('object').create_image()

# The image to use when the editor is disabled:
inactive_image = ImageResource('inactive').create_image()

# String types:
string_type = (str, unicode)

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class SimpleEditor(Editor):
    """ Simply style of editor for a drag-and-drop editor, which is both a drag
        source and a drop target.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the editor a drop target?
    drop_target = Bool(True)

    # Is the editor a drag source?
    drag_source = Bool(True)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        # Determine the drag/drop type:
        value = self.value
        self._is_list = isinstance(value, list)
        self._is_file = (isinstance(value, string_type) or
                         (self._is_list and (len(value) > 0) and
                          isinstance(value[0], string_type)))

        # Get the right image to use:
        image = self.factory.image
        if image is not None:
            image = image.create_image()
            disabled_image = self.factory.disabled_image
            if disabled_image is not None:
                disabled_image = disabled_image.create_image()
        else:
            disabled_image = inactive_image
            image = object_image
            if self._is_file:
                image = file_image

        self._image = image.ConvertToBitmap()
        if disabled_image is not None:
            self._disabled_image = disabled_image.ConvertToBitmap()
        else:
            data = numpy.reshape(numpy.fromstring(image.GetData(),
                                                  numpy.uint8),
                                 (-1, 3)) * numpy.array([[0.297, 0.589, 0.114]])
            g = data[:, 0] + data[:, 1] + data[:, 2]
            data[:, 0] = data[:, 1] = data[:, 2] = g
            image.SetData(numpy.ravel(data.astype(numpy.uint8)).tostring())
            image.SetMaskColour(0, 0, 0)
            self._disabled_image = image.ConvertToBitmap()

        # Create the control and set up the event handlers:
        self.control = control = wx.Window(
            parent, -1, size=wx.Size(image.GetWidth(), image.GetHeight()))
        self.set_tooltip()

        if self.drop_target:
            control.SetDropTarget(PythonDropTarget(self))

        wx.EVT_LEFT_DOWN(control, self._left_down)
        wx.EVT_LEFT_UP(control, self._left_up)
        wx.EVT_MOTION(control, self._mouse_move)
        wx.EVT_PAINT(control, self._on_paint)

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        control = self.control
        wx.EVT_LEFT_DOWN(control, None)
        wx.EVT_LEFT_UP(control, None)
        wx.EVT_MOTION(control, None)
        wx.EVT_PAINT(control, None)

        super(SimpleEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        return

#-- Private Methods ------------------------------------------------------

    #-------------------------------------------------------------------------
    #  Returns the processed version of a drag request's data:
    #-------------------------------------------------------------------------

    def _get_drag_data(self, data):
        """ Returns the processed version of a drag request's data.
        """
        if isinstance(data, list):

            if Binding is not None and isinstance(data[0], Binding):
                data = [item.obj for item in data]

            if File is not None and isinstance(data[0], File):
                data = [item.absolute_path for item in data]
                if not self._is_file:
                    result = []
                    for file in data:
                        item = self._unpickle(file)
                        if item is not None:
                            result.append(item)
                    data = result

        else:
            if Binding is not None and isinstance(data, Binding):
                data = data.obj

            if File is not None and isinstance(data, File):
                data = data.absolute_path
                if not self._is_file:
                    object = self._unpickle(data)
                    if object is not None:
                        data = object

        return data

    #-------------------------------------------------------------------------
    #  Returns the unpickled version of a specified file (if possible):
    #-------------------------------------------------------------------------

    def _unpickle(self, file_name):
        """ Returns the unpickled version of a specified file (if possible).
        """
        fh = None
        try:
            fh = open(file_name, 'rb')
            object = load(fh)
        except:
            object = None

        if fh is not None:
            fh.close()

        return object

#-- wxPython Event Handlers ----------------------------------------------

    def _on_paint(self, event):
        """ Called when the control needs repainting.
        """
        image = self._image
        control = self.control
        if not control.IsEnabled():
            image = self._disabled_image

        wdx, wdy = control.GetClientSizeTuple()
        wx.PaintDC(control).DrawBitmap(
            image,
            (wdx - image.GetWidth()) / 2,
            (wdy - image.GetHeight()) / 2,
            True)

    def _left_down(self, event):
        """ Handles the left mouse button being pressed.
        """
        if self.control.IsEnabled() and self.drag_source:
            self._x, self._y = event.GetX(), event.GetY()
            self.control.CaptureMouse()

        event.Skip()

    def _left_up(self, event):
        """ Handles the left mouse button being released.
        """
        if self._x is not None:
            self._x = None
            self.control.ReleaseMouse()

        event.Skip()

    def _mouse_move(self, event):
        """ Handles the mouse being moved.
        """
        if self._x is not None:
            if ((abs(self._x - event.GetX()) +
                 abs(self._y - event.GetY())) >= 3):
                self.control.ReleaseMouse()
                self._x = None
                if self._is_file:
                    FileDropSource(self.control, self.value)
                else:
                    PythonDropSource(self.control, self.value)

        event.Skip()

#----- Drag and drop event handlers: -------------------------------------

    #-------------------------------------------------------------------------
    #  Handles a Python object being dropped on the control:
    #-------------------------------------------------------------------------

    def wx_dropped_on(self, x, y, data, drag_result):
        """ Handles a Python object being dropped on the tree.
        """
        try:
            self.value = self._get_drag_data(data)
            return drag_result
        except:
            return wx.DragNone

    #-------------------------------------------------------------------------
    #  Handles a Python object being dragged over the control:
    #-------------------------------------------------------------------------

    def wx_drag_over(self, x, y, data, drag_result):
        """ Handles a Python object being dragged over the tree.
        """
        try:
            self.object.base_trait(
                self.name).validate(
                self.object,
                self.name,
                self._get_drag_data(data))
            return drag_result
        except:
            return wx.DragNone

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(SimpleEditor):
    """ Custom style of drag-and-drop editor, which is not a drag source.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the editor a drag source? This value overrides the default.
    drag_source = False

#-------------------------------------------------------------------------
#  'ReadonlyEditor' class:
#-------------------------------------------------------------------------


class ReadonlyEditor(SimpleEditor):
    """ Read-only style of drag-and-drop editor, which is not a drop target.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the editor a drop target? This value overrides the default.
    drop_target = False

#-------------------------------------------------------------------------
#  'FileDropSource' class:
#-------------------------------------------------------------------------


class FileDropSource(wx.DropSource):
    """ Represents a draggable file.
    """
    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, source, files):
        """ Initializes the object.
        """
        self.handler = None
        self.allow_move = True

        # Put the data to be dragged on the clipboard:
        clipboard.data = files
        clipboard.source = source
        clipboard.drop_source = self

        data_object = wx.FileDataObject()
        if isinstance(files, string_type):
            files = [files]

        for file in files:
            data_object.AddFile(file)

        # Create the drop source and begin the drag and drop operation:
        super(FileDropSource, self).__init__(source)
        self.SetData(data_object)
        self.result = self.DoDragDrop(True)

    #-------------------------------------------------------------------------
    #  Called when the data has been dropped:
    #-------------------------------------------------------------------------

    def on_dropped(self, drag_result):
        """ Called when the data has been dropped. """
        return

## EOF ########################################################################
